import dataclasses
import typing
import unittest.mock

import pytest

import sila
from unitelabs.cdk.sila.command.intermediate import Intermediate
from unitelabs.cdk.sila.command.status import Status
from unitelabs.cdk.sila.data_types.any import Any
from unitelabs.cdk.sila.utils.docstring import Docstring, parse_docstring


class TestGoogleStyle:
    def test_should_parse_function_docstring(self):
        # Create function
        def test_function(
            param_a: str, param_b: int, *, status: Status, intermediate: Intermediate[tuple[bytes, int]]
        ) -> tuple[bool, float]:
            """
            Some short introduction.

            With more details later on.

            Examples:
              Example on how to use the function:
              >>> for i in test_function("a", 1):
              ...     print(i)
              True
              1.23

            Args:
              ParamA: The first parameter.
              DifferentB: The second parameter.

            Yields:
              InterResponseA: The first intermediate response.
              InterResponseB: The second intermediate response.

            Returns:
              ResponseA: The first response.
              ResponseB: The second response.

            Raises:
              RuntimeError: If the matrix is not numerically invertible.
            """

            return True, 1.23

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description=(
                "Some short introduction.\n\nWith more details later on.\n\nExample on how to use the function:\n"
                '>>> for i in test_function("a", 1):\n...     print(i)\nTrue\n1.23\n'
            ),
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first parameter.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="DifferentB",
                    display_name="Different B",
                    description="The second parameter.",
                    data_type=sila.Integer,
                ),
            },
            yields={
                "inter_response_a": sila.Element(
                    identifier="InterResponseA",
                    display_name="Inter Response A",
                    description="The first intermediate response.",
                    data_type=sila.Binary,
                ),
                "inter_response_b": sila.Element(
                    identifier="InterResponseB",
                    display_name="Inter Response B",
                    description="The second intermediate response.",
                    data_type=sila.Integer,
                ),
            },
            returns={
                "response_a": sila.Element(
                    identifier="ResponseA",
                    display_name="Response A",
                    description="The first response.",
                    data_type=sila.Boolean,
                ),
                "response_b": sila.Element(
                    identifier="ResponseB",
                    display_name="Response B",
                    description="The second response.",
                    data_type=sila.Real,
                ),
            },
            raises={
                "runtime_error": docstring.raises["runtime_error"],
            },
        )

        assert issubclass(docstring.raises["runtime_error"], sila.DefinedExecutionError)
        assert docstring.raises["runtime_error"].identifier == "RuntimeError"
        assert docstring.raises["runtime_error"].display_name == "Runtime Error"
        assert docstring.raises["runtime_error"].description == "If the matrix is not numerically invertible."

    def test_should_parse_class_docstring(self):
        # Create class
        class TestClass:
            """
            Some short introduction.

            With more details later on.

            Examples:
              Example on how to use the class:
              >>> test_class = TestClass("a", 1)
              ... print(test_class.param_a)
              a

            Attributes:
              ParamA: The first value.
              ParamB: The second value.
            """

            param_a: str
            param_b: int

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(TestClass, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description=(
                "Some short introduction.\n\nWith more details later on.\n\nExample on how to use the class:\n"
                '>>> test_class = TestClass("a", 1)\n... print(test_class.param_a)\na\n'
            ),
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first value.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The second value.",
                    data_type=sila.Integer,
                ),
            },
        )

    def test_should_parse_dataclass_docstring(self):
        # Create dataclass
        @dataclasses.dataclass
        class TestClass:
            """
            Some short introduction.

            With more details later on.

            Examples:
              Example on how to use the class:
              >>> test_class = TestClass("a", 1)
              ... print(test_class.param_a)
              a

            Attributes:
              ParamA: The first value.
              ParamB: The second value.
            """

            param_a: str
            param_b: int

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(TestClass, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description=(
                "Some short introduction.\n\nWith more details later on.\n\nExample on how to use the class:\n"
                '>>> test_class = TestClass("a", 1)\n... print(test_class.param_a)\na\n'
            ),
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first value.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The second value.",
                    data_type=sila.Integer,
                ),
            },
        )

    def test_should_parse_typed_dict_docstring(self):
        # Create typed dict
        class TestClass(typing.TypedDict):
            """
            Some short introduction.

            With more details later on.

            Examples:
              Example on how to use the class:
              >>> test_class = TestClass("a", 1)
              ... print(test_class.param_a)
              a

            Attributes:
              ParamA: The first value.
              ParamB: The second value.
            """

            param_a: str
            param_b: int

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(TestClass, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description=(
                "Some short introduction.\n\nWith more details later on.\n\nExample on how to use the class:\n"
                '>>> test_class = TestClass("a", 1)\n... print(test_class.param_a)\na\n'
            ),
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first value.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The second value.",
                    data_type=sila.Integer,
                ),
            },
        )

    def test_should_parse_description(self):
        # Create function
        def test_function() -> None:
            """Test function."""

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(description="Test function.")

    def test_should_parse_multiline_description(self):
        # Create function
        def test_function() -> None:
            """
            Test function.

            With more details later on.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(description="Test function.\n\nWith more details later on.")

    def test_should_parse_missing_parameter(self):
        # Create function
        def test_function(param_a) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            parameters={"param_a": sila.Element(identifier="ParamA", display_name="Param A", data_type=Any)}
        )

    def test_should_parse_no_parameter(self):
        # Create function
        def test_function() -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(parameters={})

    def test_should_parse_one_parameter(self):
        # Create function
        def test_function(param_a: bool) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            parameters={"param_a": sila.Element(identifier="ParamA", display_name="Param A", data_type=sila.Boolean)}
        )

    def test_should_parse_one_parameter_with_description(self):
        # Create function
        def test_function(param_a: bool) -> None:
            """Test function.

            Args:
              ParamA: The parameter.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA", display_name="Param A", description="The parameter.", data_type=sila.Boolean
                )
            },
        )

    def test_should_parse_one_parameter_with_multiline_description(self):
        # Create function
        def test_function(param_a: bool) -> None:
            """Test function.

            Args:
              ParamA: The parameter
                over multiple
                lines.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The parameter\nover multiple\nlines.",
                    data_type=sila.Boolean,
                )
            },
        )

    def test_should_parse_one_parameter_with_annotation(self):
        # Create function
        def test_function(param_a) -> None:
            """Test function.

            Args:
              ParamA (int): The parameter.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA", display_name="Param A", description="The parameter.", data_type=Any
                )
            },
        )

    def test_should_parse_multiple_parameter(self):
        # Create function
        def test_function(param_a: str, param_b: int, param_c: float) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            parameters={
                "param_a": sila.Element(identifier="ParamA", display_name="Param A", data_type=sila.String),
                "param_b": sila.Element(identifier="ParamB", display_name="Param B", data_type=sila.Integer),
                "param_c": sila.Element(identifier="ParamC", display_name="Param C", data_type=sila.Real),
            }
        )

    def test_should_parse_multiple_parameter_with_description(self):
        # Create function
        def test_function(param_a: str, param_b: int, param_c: float) -> None:
            """Test function.

            Args:
              ParamA: The string parameter.
              ParamB: The integer parameter.
              ParamC: The float parameter.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The string parameter.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The integer parameter.",
                    data_type=sila.Integer,
                ),
                "param_c": sila.Element(
                    identifier="ParamC", display_name="Param C", description="The float parameter.", data_type=sila.Real
                ),
            },
        )

    def test_should_parse_multiple_parameter_with_annotation(self):
        # Create function
        def test_function(param_a: str, param_b, param_c: float) -> None:
            """Test function.

            Args:
              ParamA (bool): The string parameter.
              ParamB (bool): The integer parameter.
              ParamC: The float parameter.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The string parameter.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The integer parameter.",
                    data_type=Any,
                ),
                "param_c": sila.Element(
                    identifier="ParamC", display_name="Param C", description="The float parameter.", data_type=sila.Real
                ),
            },
        )

    def test_should_skip_missing_parameter_description(self):
        # Create function
        def test_function(param_a: str, param_b: int, param_c: float) -> None:
            """Test function.

            Args:
              ParamA: The string parameter.
              ParamB: The integer parameter.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The string parameter.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The integer parameter.",
                    data_type=sila.Integer,
                ),
                "param_c": sila.Element(identifier="ParamC", display_name="Param C", data_type=sila.Real),
            },
        )

    def test_should_ignore_untyped_parameter_description(self):
        # Create function
        def test_function(param_a: str, param_b: int, param_c: float) -> None:
            """Test function.

            Args:
              ParamA: The string parameter.
              ParamB: The integer parameter.
              ParamC: The float parameter.
              ParamD: The boolean parameter.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The string parameter.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="ParamB",
                    display_name="Param B",
                    description="The integer parameter.",
                    data_type=sila.Integer,
                ),
                "param_c": sila.Element(
                    identifier="ParamC", display_name="Param C", description="The float parameter.", data_type=sila.Real
                ),
            },
        )

    def test_should_parse_missing_return(self):
        # Create function
        def test_function(): ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            returns={"response_0": sila.Element(identifier="Response0", display_name="Response 0", data_type=Any)}
        )

    def test_should_parse_no_return(self):
        # Create function
        def test_function() -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(returns={})

    def test_should_parse_one_return(self):
        # Create function
        def test_function() -> bool: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            returns={
                "response_0": sila.Element(identifier="Response0", display_name="Response 0", data_type=sila.Boolean)
            }
        )

    def test_should_parse_one_return_with_description(self):
        # Create function
        def test_function() -> bool:
            """Test function.

            Returns:
              The response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_0": sila.Element(
                    identifier="Response0",
                    display_name="Response 0",
                    description="The response.",
                    data_type=sila.Boolean,
                )
            },
        )

    def test_should_parse_one_return_with_multiline_description(self):
        # Create function
        def test_function() -> bool:
            """Test function.

            Returns:
              The response
                over multiple
                lines.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_0": sila.Element(
                    identifier="Response0",
                    display_name="Response 0",
                    description="The response\nover multiple\nlines.",
                    data_type=sila.Boolean,
                )
            },
        )

    def test_should_parse_one_return_with_name(self):
        # Create function
        def test_function() -> bool:
            """Test function.

            Returns:
              Response: The response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response": sila.Element(
                    identifier="Response", display_name="Response", description="The response.", data_type=sila.Boolean
                )
            },
        )

    def test_should_parse_one_return_with_annotation(self):
        # Create function
        def test_function():
            """Test function.

            Returns:
              Response (int): The response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response": sila.Element(
                    identifier="Response",
                    display_name="Response",
                    description="The response.",
                    data_type=Any,
                )
            },
        )

    def test_should_parse_multiple_return(self):
        # Create function
        def test_function() -> tuple[str, int, float]: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            returns={
                "response_0": sila.Element(identifier="Response0", display_name="Response 0", data_type=sila.String),
                "response_1": sila.Element(identifier="Response1", display_name="Response 1", data_type=sila.Integer),
                "response_2": sila.Element(identifier="Response2", display_name="Response 2", data_type=sila.Real),
            }
        )

    def test_should_parse_multiple_return_with_description(self):
        # Create function
        def test_function() -> tuple[str, int, float]:
            """Test function.

            Returns:
              The string response.
              The integer response
                over multiple lines.
              The float response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_0": sila.Element(
                    identifier="Response0",
                    display_name="Response 0",
                    data_type=sila.String,
                    description="The string response.",
                ),
                "response_1": sila.Element(
                    identifier="Response1",
                    display_name="Response 1",
                    data_type=sila.Integer,
                    description="The integer response\nover multiple lines.",
                ),
                "response_2": sila.Element(
                    identifier="Response2",
                    display_name="Response 2",
                    data_type=sila.Real,
                    description="The float response.",
                ),
            },
        )

    def test_should_parse_multiple_return_with_name(self):
        # Create function
        def test_function() -> tuple[str, int, float]:
            """Test function.

            Returns:
              ResponseString: The string response.
              ResponseInt: The integer response
                over multiple lines.
              ResponseFloat: The float response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    description="The string response.",
                    data_type=sila.String,
                ),
                "response_int": sila.Element(
                    identifier="ResponseInt",
                    display_name="Response Int",
                    description="The integer response\nover multiple lines.",
                    data_type=sila.Integer,
                ),
                "response_float": sila.Element(
                    identifier="ResponseFloat",
                    display_name="Response Float",
                    description="The float response.",
                    data_type=sila.Real,
                ),
            },
        )

    def test_should_parse_multiple_return_with_annotation(self):
        # Create function
        def test_function() -> tuple[str, int, float]:
            """Test function.

            Returns:
              ResponseString: The string response.
              ResponseInt (bool): The integer response
                over multiple lines.
              ResponseFloat: The float response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    description="The string response.",
                    data_type=sila.String,
                ),
                "response_int": sila.Element(
                    identifier="ResponseInt",
                    display_name="Response Int",
                    description="The integer response\nover multiple lines.",
                    data_type=sila.Integer,
                ),
                "response_float": sila.Element(
                    identifier="ResponseFloat",
                    display_name="Response Float",
                    description="The float response.",
                    data_type=sila.Real,
                ),
            },
        )

    def test_should_skip_missing_return_description(self):
        # Create function
        def test_function() -> tuple[str, int, float]:
            """Test function.

            Returns:
              ResponseString: The string response.
              The integer response
                over multiple lines.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    description="The string response.",
                    data_type=sila.String,
                ),
                "response_1": sila.Element(
                    identifier="Response1",
                    display_name="Response 1",
                    description="The integer response\nover multiple lines.",
                    data_type=sila.Integer,
                ),
                "response_2": sila.Element(
                    identifier="Response2",
                    display_name="Response 2",
                    data_type=sila.Real,
                ),
            },
        )

    def test_should_ignore_untyped_return_description(self):
        # Create function
        def test_function() -> tuple[str, int, float]:
            """Test function.

            Returns:
              ResponseString: The string response.
              ResponseInt: The integer response
                over multiple lines.
              ResponseFloat: The float response.
              ResponseBoolean: The boolean response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            returns={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    description="The string response.",
                    data_type=sila.String,
                ),
                "response_int": sila.Element(
                    identifier="ResponseInt",
                    display_name="Response Int",
                    description="The integer response\nover multiple lines.",
                    data_type=sila.Integer,
                ),
                "response_float": sila.Element(
                    identifier="ResponseFloat",
                    display_name="Response Float",
                    description="The float response.",
                    data_type=sila.Real,
                ),
            },
        )

    def test_should_parse_missing_yield(self):
        # Create function
        def test_function(intermediate: Intermediate) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            yields={
                "intermediate_response_0": sila.Element(
                    identifier="IntermediateResponse0",
                    display_name="Intermediate Response 0",
                    data_type=Any,
                )
            }
        )

    def test_should_parse_no_yield(self):
        # Create function
        def test_function(intermediate: Intermediate[None]) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(yields={})

    def test_should_parse_one_yield(self):
        # Create function
        def test_function(intermediate: Intermediate[bool]) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            yields={
                "intermediate_response_0": sila.Element(
                    identifier="IntermediateResponse0",
                    display_name="Intermediate Response 0",
                    data_type=sila.Boolean,
                )
            }
        )

    def test_should_parse_one_yield_with_description(self):
        # Create function
        def test_function(intermediate: Intermediate[bool]) -> None:
            """Test function.

            Yields:
              The intermediate response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "intermediate_response_0": sila.Element(
                    identifier="IntermediateResponse0",
                    display_name="Intermediate Response 0",
                    data_type=sila.Boolean,
                    description="The intermediate response.",
                )
            },
        )

    def test_should_parse_one_yield_with_multiline_description(self):
        # Create function
        def test_function(intermediate: Intermediate[bool]) -> None:
            """Test function.

            Yields:
              The intermediate response
                over multiple
                lines.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "intermediate_response_0": sila.Element(
                    identifier="IntermediateResponse0",
                    display_name="Intermediate Response 0",
                    data_type=sila.Boolean,
                    description="The intermediate response\nover multiple\nlines.",
                )
            },
        )

    def test_should_parse_one_yield_with_name(self):
        # Create function
        def test_function(intermediate: Intermediate[bool]) -> None:
            """Test function.

            Yields:
              IntermediateResponse: The intermediate response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "intermediate_response": sila.Element(
                    identifier="IntermediateResponse",
                    display_name="Intermediate Response",
                    data_type=sila.Boolean,
                    description="The intermediate response.",
                )
            },
        )

    def test_should_parse_one_yield_with_annotation(self):
        # Create function
        def test_function(intermediate: Intermediate) -> None:
            """Test function.

            Yields:
              IntermediateResponse (int): The intermediate response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "intermediate_response": sila.Element(
                    identifier="IntermediateResponse",
                    display_name="Intermediate Response",
                    data_type=Any,
                    description="The intermediate response.",
                )
            },
        )

    def test_should_parse_multiple_yields(self):
        # Create function
        def test_function(intermediate: Intermediate[tuple[str, int, float]]) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            yields={
                "intermediate_response_0": sila.Element(
                    identifier="IntermediateResponse0",
                    display_name="Intermediate Response 0",
                    data_type=sila.String,
                ),
                "intermediate_response_1": sila.Element(
                    identifier="IntermediateResponse1",
                    display_name="Intermediate Response 1",
                    data_type=sila.Integer,
                ),
                "intermediate_response_2": sila.Element(
                    identifier="IntermediateResponse2",
                    display_name="Intermediate Response 2",
                    data_type=sila.Real,
                ),
            }
        )

    def test_should_parse_multiple_yield_with_description(self):
        # Create function
        def test_function(intermediate: Intermediate[tuple[str, int, float]]) -> None:
            """Test function.

            Yields:
              The string response.
              The integer response
                over multiple lines.
              The float response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "intermediate_response_0": sila.Element(
                    identifier="IntermediateResponse0",
                    display_name="Intermediate Response 0",
                    data_type=sila.String,
                    description="The string response.",
                ),
                "intermediate_response_1": sila.Element(
                    identifier="IntermediateResponse1",
                    display_name="Intermediate Response 1",
                    data_type=sila.Integer,
                    description="The integer response\nover multiple lines.",
                ),
                "intermediate_response_2": sila.Element(
                    identifier="IntermediateResponse2",
                    display_name="Intermediate Response 2",
                    data_type=sila.Real,
                    description="The float response.",
                ),
            },
        )

    def test_should_parse_multiple_yield_with_name(self):
        # Create function
        def test_function(intermediate: Intermediate[tuple[str, int, float]]) -> None:
            """Test function.

            Yields:
              ResponseString: The string response.
              ResponseInt: The integer response
                over multiple lines.
              ResponseFloat: The float response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    data_type=sila.String,
                    description="The string response.",
                ),
                "response_int": sila.Element(
                    identifier="ResponseInt",
                    display_name="Response Int",
                    data_type=sila.Integer,
                    description="The integer response\nover multiple lines.",
                ),
                "response_float": sila.Element(
                    identifier="ResponseFloat",
                    display_name="Response Float",
                    data_type=sila.Real,
                    description="The float response.",
                ),
            },
        )

    def test_should_parse_multiple_yields_with_annotation(self):
        # Create function
        def test_function(intermediate: Intermediate[tuple[str, int, float]]) -> None:
            """Test function.

            Yields:
              ResponseString: The string response.
              ResponseInt (bool): The integer response
                over multiple lines.
              ResponseFloat: The float response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    description="The string response.",
                    data_type=sila.String,
                ),
                "response_int": sila.Element(
                    identifier="ResponseInt",
                    display_name="Response Int",
                    description="The integer response\nover multiple lines.",
                    data_type=sila.Integer,
                ),
                "response_float": sila.Element(
                    identifier="ResponseFloat",
                    display_name="Response Float",
                    description="The float response.",
                    data_type=sila.Real,
                ),
            },
        )

    def test_should_skip_missing_yield_description(self):
        # Create function
        def test_function(intermediate: Intermediate[tuple[str, int, float]]) -> None:
            """Test function.

            Yields:
              ResponseString: The string response.
              The integer response
                over multiple lines.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    data_type=sila.String,
                    description="The string response.",
                ),
                "intermediate_response_1": sila.Element(
                    identifier="IntermediateResponse1",
                    display_name="Intermediate Response 1",
                    data_type=sila.Integer,
                    description="The integer response\nover multiple lines.",
                ),
                "intermediate_response_2": sila.Element(
                    identifier="IntermediateResponse2",
                    display_name="Intermediate Response 2",
                    data_type=sila.Real,
                ),
            },
        )

    def test_should_ignore_untyped_yield_description(self):
        # Create function
        def test_function(intermediate: Intermediate[tuple[str, int, float]]) -> None:
            """Test function.

            Yields:
              ResponseString: The string response.
              ResponseInt: The integer response
                over multiple lines.
              ResponseFloat: The float response.
              ResponseBoolean: The boolean response.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns(UserWarning):
            docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Test function.",
            yields={
                "response_string": sila.Element(
                    identifier="ResponseString",
                    display_name="Response String",
                    data_type=sila.String,
                    description="The string response.",
                ),
                "response_int": sila.Element(
                    identifier="ResponseInt",
                    display_name="Response Int",
                    data_type=sila.Integer,
                    description="The integer response\nover multiple lines.",
                ),
                "response_float": sila.Element(
                    identifier="ResponseFloat",
                    display_name="Response Float",
                    data_type=sila.Real,
                    description="The float response.",
                ),
            },
        )

    def test_should_ignore_self(self):
        # Create function
        def test_function(self, param_a: str) -> None: ...

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            parameters={"param_a": sila.Element(identifier="ParamA", display_name="Param A", data_type=sila.String)}
        )

    def test_should_keep_casing(self):
        # Create function
        def test_function(uuid: int, prefix_uuid: int, uuid_suffix: int, prefix_uuid_suffix: int) -> None:
            """
            Some short introduction.

            Args:
              UUID: The UUID.
              PrefixUUID: The prefix UUID.
              UUIDSuffix: The UUID suffix.
              PrefixUUIDSuffix: The prefix UUID suffix.
            """

        # Parse docstring
        feature = unittest.mock.Mock()
        docstring = parse_docstring(test_function, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Some short introduction.",
            parameters={
                "uuid": sila.Element(
                    identifier="UUID", display_name="UUID", description="The UUID.", data_type=sila.Integer
                ),
                "prefix_uuid": sila.Element(
                    identifier="PrefixUUID",
                    display_name="Prefix UUID",
                    description="The prefix UUID.",
                    data_type=sila.Integer,
                ),
                "uuid_suffix": sila.Element(
                    identifier="UUIDSuffix",
                    display_name="UUID Suffix",
                    description="The UUID suffix.",
                    data_type=sila.Integer,
                ),
                "prefix_uuid_suffix": sila.Element(
                    identifier="PrefixUUIDSuffix",
                    display_name="Prefix UUID Suffix",
                    description="The prefix UUID suffix.",
                    data_type=sila.Integer,
                ),
            },
        )


class TestRestructured:
    def test_should_parse_docstring(self):
        # Create function
        def test_function(
            param_a: str, param_b: int, *, status: Status, intermediate: Intermediate[tuple[bytes, int]]
        ) -> tuple[bool, float]:
            """
            Some short introduction.

            With more details later on.

            .. parameter:: The first parameter.
            .. parameter:: The second parameter.
              :name: Alternative Parameter
            .. yield:: The first intermediate response.
              :name: Inter Response A
            .. yield:: The second intermediate response.
              :name: Inter Response B
            .. return:: The first response.
              :name: Response A
            .. return:: The second response.
              :name: Response B
            """

            return True, 1.23

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns():
            docstring = parse_docstring(test_function, feature)

        assert docstring == Docstring(
            description="Some short introduction.\n\nWith more details later on.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first parameter.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="AlternativeParameter",
                    display_name="Alternative Parameter",
                    description="The second parameter.",
                    data_type=sila.Integer,
                ),
            },
            yields={
                "inter_response_a": sila.Element(
                    identifier="InterResponseA",
                    display_name="Inter Response A",
                    description="The first intermediate response.",
                    data_type=sila.Binary,
                ),
                "inter_response_b": sila.Element(
                    identifier="InterResponseB",
                    display_name="Inter Response B",
                    description="The second intermediate response.",
                    data_type=sila.Integer,
                ),
            },
            returns={
                "response_a": sila.Element(
                    identifier="ResponseA",
                    display_name="Response A",
                    description="The first response.",
                    data_type=sila.Boolean,
                ),
                "response_b": sila.Element(
                    identifier="ResponseB",
                    display_name="Response B",
                    description="The second response.",
                    data_type=sila.Real,
                ),
            },
            raises={},
        )

    def test_should_parse_class_docstring(self):
        # Create class
        class TestClass:
            """
            Some short introduction.

            With more details later on.

            .. parameter:: The first value.
            .. parameter:: The second value.
              :name: Alternative Value
            """

            param_a: str
            param_b: int

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns():
            docstring = parse_docstring(TestClass, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Some short introduction.\n\nWith more details later on.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first value.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="AlternativeValue",
                    display_name="Alternative Value",
                    description="The second value.",
                    data_type=sila.Integer,
                ),
            },
        )

    def test_should_parse_dataclass_docstring(self):
        # Create class
        @dataclasses.dataclass
        class TestClass:
            """
            Some short introduction.

            With more details later on.

            .. parameter:: The first value.
            .. parameter:: The second value.
              :name: Alternative Value
            """

            param_a: str
            param_b: int

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns():
            docstring = parse_docstring(TestClass, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Some short introduction.\n\nWith more details later on.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first value.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="AlternativeValue",
                    display_name="Alternative Value",
                    description="The second value.",
                    data_type=sila.Integer,
                ),
            },
        )

    def test_should_parse_typed_dict_docstring(self):
        # Create class
        class TestClass(typing.TypedDict):
            """
            Some short introduction.

            With more details later on.

            .. parameter:: The first value.
            .. parameter:: The second value.
              :name: Alternative Value
            """

            param_a: str
            param_b: int

        # Parse docstring
        feature = unittest.mock.Mock()
        with pytest.warns():
            docstring = parse_docstring(TestClass, feature)

        # Assert that the method returns the correct value
        assert docstring == Docstring(
            description="Some short introduction.\n\nWith more details later on.",
            parameters={
                "param_a": sila.Element(
                    identifier="ParamA",
                    display_name="Param A",
                    description="The first value.",
                    data_type=sila.String,
                ),
                "param_b": sila.Element(
                    identifier="AlternativeValue",
                    display_name="Alternative Value",
                    description="The second value.",
                    data_type=sila.Integer,
                ),
            },
        )



import json
import pathlib
import typing
import uuid

import pytest
from click.testing import CliRunner

from unitelabs.cdk.cli.certificate import MutuallyExclusiveOptions
from unitelabs.cdk.config import ConnectorBaseConfig


@pytest.fixture
def create_temp_config(tmp_path) -> typing.Callable[[str], tuple[ConnectorBaseConfig, pathlib.Path]]:
    def config(ext: str) -> tuple[ConnectorBaseConfig, pathlib.Path]:
        config_path = tmp_path / f"config.{ext}"
        config = ConnectorBaseConfig()
        config.dump(config_path)
        return (config, config_path)

    return config


@pytest.fixture
def temp_config_with_comment(
    create_temp_config,
) -> typing.Generator[tuple[ConnectorBaseConfig, pathlib.Path], None, None]:
    config, config_path = create_temp_config("yaml")
    with config_path.open("a") as f:
        f.write("# comment\n")
    yield (config, config_path)


class TestGenerate:
    @pytest.mark.parametrize("ext", ["json", "yaml", "yml"])
    def test_should_accept_config_file(self, create_temp_config, ext):
        from unitelabs.cdk.cli import certificate

        runner = CliRunner()
        config, config_path = create_temp_config(ext)

        result = runner.invoke(
            certificate, ["generate", "--config-path", config_path, "--target", config_path.parent], input="n"
        )
        assert result.exit_code == 0

        config_file_contents = ConnectorBaseConfig.load(config_path)

        assert config_file_contents == config

    @pytest.mark.parametrize("ext", ["json", "yaml", "yml"])
    def test_should_update_config_file_with_prompt(self, create_temp_config, ext):
        from unitelabs.cdk.cli import certificate

        runner = CliRunner()
        config, config_path = create_temp_config(ext)

        result = runner.invoke(
            certificate, ["generate", "--config-path", config_path, "--target", config_path.parent], input="y"
        )
        assert result.exit_code == 0

        config_file_contents = ConnectorBaseConfig.load(config_path)

        # cert, key, and tls values have been updated
        assert config_file_contents.sila_server
        assert config_file_contents.sila_server.tls
        assert isinstance(config_file_contents.sila_server.certificate_chain, bytes)
        assert isinstance(config_file_contents.sila_server.private_key, bytes)

        # all other values are the same
        assert config_file_contents != config

        config_file_contents.sila_server.certificate_chain = None
        config_file_contents.sila_server.private_key = None
        config_file_contents.sila_server.tls = False

        assert config_file_contents == config

    @pytest.mark.parametrize("ext", ["json", "yaml", "yml"])
    def test_update_should_suppress_prompt_and_update_config(self, create_temp_config, ext):
        from unitelabs.cdk.cli import certificate

        runner = CliRunner()
        config, config_path = create_temp_config(ext)

        result = runner.invoke(
            certificate, ["generate", "--config-path", config_path, "--target", config_path.parent, "-y"]
        )
        assert result.exit_code == 0

        config_file_contents = ConnectorBaseConfig.load(config_path)

        # cert, key, and tls values have been updated
        assert config_file_contents.sila_server
        assert config_file_contents.sila_server.tls
        assert isinstance(config_file_contents.sila_server.certificate_chain, bytes)
        assert isinstance(config_file_contents.sila_server.private_key, bytes)

        # all other values are the same
        assert config_file_contents != config

        config_file_contents.sila_server.certificate_chain = None
        config_file_contents.sila_server.private_key = None
        config_file_contents.sila_server.tls = False

        assert config_file_contents == config

    def test_uuid_and_host_should_generate_certs(self, tmp_path):
        from unitelabs.cdk.cli import certificate

        cert_path = tmp_path / "cert.pem"
        key_path = tmp_path / "key.pem"

        assert not cert_path.exists()
        assert not key_path.exists()

        runner = CliRunner()
        result = runner.invoke(
            certificate, ["generate", "--uuid", str(uuid.uuid4()), "--host", "localhost", "--target", tmp_path]
        )
        assert result.exit_code == 0

        assert cert_path.exists()
        assert key_path.exists()

    def test_should_not_allow_config_path_and_uuid(self, create_temp_config):
        from unitelabs.cdk.cli import certificate

        _, config_path = create_temp_config("json")

        runner = CliRunner()
        result = runner.invoke(certificate, ["generate", "--config-path", config_path, "--uuid", str(uuid.uuid4())])

        assert result.exit_code == 1
        assert result.exc_info
        assert result.exc_info[0] == MutuallyExclusiveOptions

    def test_should_not_allow_config_path_and_host(self, create_temp_config):
        from unitelabs.cdk.cli import certificate

        _, config_path = create_temp_config("json")

        runner = CliRunner()
        result = runner.invoke(certificate, ["generate", "--config-path", config_path, "--host", "localhost"])

        assert result.exit_code == 1
        assert result.exc_info
        assert result.exc_info[0] == MutuallyExclusiveOptions

    def test_should_not_allow_config_path_and_uuid_and_host(self, create_temp_config):
        from unitelabs.cdk.cli import certificate

        _, config_path = create_temp_config("json")

        runner = CliRunner()
        result = runner.invoke(
            certificate, ["generate", "--config-path", config_path, "--uuid", str(uuid.uuid4()), "--host", "localhost"]
        )

        assert result.exit_code == 1
        assert result.exc_info
        assert result.exc_info[0] == MutuallyExclusiveOptions

    def test_should_not_allow_embed_and_target(self, create_temp_config):
        from unitelabs.cdk.cli import certificate

        _, config_path = create_temp_config("json")

        runner = CliRunner()
        result = runner.invoke(
            certificate,
            [
                "generate",
                "--config-path",
                config_path,
                "--embed",
                "--target",
                config_path.parent,
            ],
        )

        assert result.exit_code == 1
        assert result.exc_info
        assert result.exc_info[0] == MutuallyExclusiveOptions

    @pytest.mark.parametrize("ext", ["json", "yaml", "yml"])
    def test_should_embed_certificate_and_key_in_config(self, create_temp_config, ext):
        from unitelabs.cdk.cli import certificate

        _, config_path = create_temp_config(ext)

        runner = CliRunner()
        result = runner.invoke(
            certificate,
            [
                "generate",
                "--config-path",
                config_path,
                "--embed",
            ],
        )
        assert result.exit_code == 0

        config_file_contents = ConnectorBaseConfig.load(config_path)

        # cert, key, and tls values have been updated
        assert config_file_contents.sila_server
        assert config_file_contents.sila_server.tls
        assert isinstance(config_file_contents.sila_server.certificate_chain, bytes)
        assert isinstance(config_file_contents.sila_server.private_key, bytes)

    def test_should_preserve_unknown_config_keys(self, tmp_path):
        from unitelabs.cdk.cli import certificate

        config_data = ConnectorBaseConfig().to_dict()
        config_data["simple"] = True

        config_path = tmp_path / "config.json"
        with config_path.open("w") as f:
            f.write(json.dumps(config_data))

        runner = CliRunner()
        result = runner.invoke(
            certificate, ["generate", "--config-path", config_path, "--target", config_path.parent, "-y"]
        )
        assert result.exit_code == 0

        with config_path.open("r") as f:
            updated = json.load(f)

        assert "simple" in updated

    def test_should_preserve_comments(self, temp_config_with_comment):
        from unitelabs.cdk.cli import certificate

        _, config_path = temp_config_with_comment

        runner = CliRunner()
        result = runner.invoke(
            certificate, ["generate", "--config-path", config_path, "--target", config_path.parent, "-y"]
        )
        assert result.exit_code == 0

        with config_path.open("r") as f:
            updated = f.read()

        assert "# comment\n" in updated



import abc
import dataclasses

import typing_extensions as typing

from unitelabs.cdk import sila


class InvalidCommandSequence(Exception):
    """The issued command breaks the sequence of commands for the device based on its role in the labware transfer."""


class LabwareNotPicked(Exception):
    """Picking up the labware item from the source device failed."""


class LabwareNotPlaced(Exception):
    """Placing the labware item at the destination device failed."""


@dataclasses.dataclass
class PositionIndex(sila.CustomDataType):
    """Specifies a position via an index number, starting at 1."""

    position_index: typing.Annotated[int, sila.constraints.MinimalInclusive(value=1)]


@dataclasses.dataclass
class HandoverPosition(sila.CustomDataType):
    """
    Representation of a possible position of a device where labware items can be handed over.

    Can contain a sub-position, e.g. for specifying a position in a
    rack.

    Attributes:
      Position: The name of the handover position (must be unique
        within the device).
      SubPosition: The index of a sub-position within a handover
        position or the number of sub-positions respectively, e.g.
        for a rack.
    """

    position: str
    sub_position: PositionIndex


class LabwareTransferManipulatorControllerBase(sila.Feature, metaclass=abc.ABCMeta):
    """
    Handle the transfer of labware items between devices.

    This feature (together with the "Labware Transfer Site Controller" feature) provides commands to trigger the
    sub-tasks of handing over a labware item, e.g. a microtiter plate or a tube, from one device to another in a
    standardized and generic way.

    For each labware transfer a defined sequence of commands has to be called on both involved devices to ensure the
    proper synchronization of all necessary transfer actions without unwanted physical interferences and to optimize
    the transfer performance regarding the execution time. Using the generic commands, labware transfers between any
    arbitrary labware handling devices can be controlled (a robot device has not necessarily to be involved).

    Generally, a labware transfer is executed between a source and a destination device, where one of them is the
    active device (executing the handover actions) and the other one is the passive device.

    The "Labware Transfer Manipulator Controller" feature is used to control the labware transfer on the side of the
    active device to hand over labware to or take over labware from a passive device, which provides the
    "Labware Transfer Site Controller" feature.

    If a device is capable to act either as the active or as the passive device of a labware transfer it must provide
    both features.

    The complete sequence of issued transfer commands on both devices is as follows:

    1. Prior to the actual labware transfer a "Prepare For Output" command is sent to the source device to execute all
       necessary actions to be ready to release a labware item (e.g. open a tray) and simultaneously a "Prepare For
       Input" command is sent to the destination device to execute all necessary actions to be ready to receive a
       labware item (e.g. position the robotic arm near the tray of the source device).
    2. When both devices have successfully finished their "Prepare For ..." command execution, the next commands are
       issued.
    3a If the source device is the active device it will receive a "Put Labware" command to execute all necessary
       actions to put the labware item into the destination device. After the transfer has been finished successfully,
       the destination device receives a "Labware Delivered" command, that triggers all actions to be done after the
       labware item has been transferred (e.g. close the opened tray).
    3b If the destination device is the active device it will receive a "Get Labware" command to execute all necessary
       actions to get the labware from the source device (e.g. gripping the labware item). After that command has been
       finished successfully, the source device receives a "Labware Removed" command, that triggers all actions to be
       done after the labware item has been transferred (e.g. close the opened tray).

    The command sequences for an active source or destination device have always to be as follows:
    - for an active source device:        PrepareForOutput - PutLabware.
    - for an active destination device:   PrepareForInput - GetLabware.

    If the commands issued by the client differ from the respective command sequences an "Invalid Command Sequence"
    error will be raised.

    To address the location, where a labware item can be handed over to or from other devices, every device must
    provide one or more uniquely named positions (handover positions) via the "Available Handover Positions" property.
    A robot (active device) should have at least one handover position for each device that it interacts with, whereas
    most passive devices will only have one handover position. In the case of a position array (e.g. a rack), the
    position within the array is specified via the sub-position of the handover position, passed as an index number.

    To address the positions within a device where the transferred labware item has to be stored at or is to be taken
    from (e.g. the storage positions inside an incubator), the internal position is specified. Each device must provide
    the number of available internal positions via the "Number Of Internal Positions" property. In the case of no
    multiple internal positions, this property as well as the "Internal Position" parameter value must be 1.

    With the "Prepare For Input" command there is also information about the labware transferred, like labware type or
    a unique labware identifier (e.g. a barcode).

    The "Intermediate Actions" parameter of the "Put Labware" and "Get Labware" commands can be used to specify commands
    that have to be executed while a labware item is transferred to avoid unnecessary movements, e.g. if a robot has to
    get a plate from a just opened tray and a lid has to be put on the plate before it will be gripped, the lid handling
    actions have to be included in the "Get Labware" actions. The intermediate actions have to be executed in the same
    order they have been specified by the "Intermediate Actions" parameter.
    The property "Available Intermediate Actions" returns a list of commands that can be included in a "Put Labware" or
    "Get Labware" command.
    """

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="instruments.labware.manipulation",
            version="1.0",
            maturity_level="Verified",
        )

    @abc.abstractmethod
    @sila.UnobservableProperty()
    async def get_available_handover_positions(self) -> list[HandoverPosition]:
        """All handover positions of the device including the number of sub-positions."""

    @abc.abstractmethod
    @sila.UnobservableProperty()
    async def get_number_of_internal_positions(
        self,
    ) -> typing.Annotated[int, sila.constraints.MinimalInclusive(value=1)]:
        """Get the number of addressable internal positions of the device."""

    @abc.abstractmethod
    @sila.UnobservableProperty()
    async def get_available_intermediate_actions(
        self,
    ) -> list[
        typing.Annotated[
            str, sila.constraints.FullyQualifiedIdentifier(value=sila.constraints.Identifier.COMMAND_IDENTIFIER)
        ]
    ]:
        """Get all commands that can be executed within a "Put Labware" or "Get Labware" command execution."""

    @abc.abstractmethod
    @sila.ObservableCommand()
    async def prepare_for_input(
        self,
        handover_position: HandoverPosition,
        internal_position: PositionIndex,
        labware_type: str,
        labware_unique_id: str,
        *,
        status: sila.Status,
    ) -> None:
        """
        Put the device into a state in which it is ready to accept new labware at the specified handover position.

        Args:
          HandoverPosition: Indicates the position where the labware will
            be handed over.
          InternalPosition: Indicates the position which the labware will
            be stored at within the device, e.g. internal storage
            positions of an incubator.
          LabwareType: Specifies the type of labware that will be handed
            over to transfer information about the labware that the
            device might need to handle it correctly.
          LabwareUniqueID: Represents the unique identification of a
            labware in the controlling system. It is assigned by the
            system and must remain unchanged during the whole process.

        Raises:
          InvalidCommandSequence: The issued command does not follow the
            sequence of commands for the device according to its role in
            the labware transfer.
        """

    @abc.abstractmethod
    @sila.ObservableCommand()
    async def prepare_for_output(
        self,
        handover_position: HandoverPosition,
        internal_position: PositionIndex,
        *,
        status: sila.Status,
    ) -> None:
        """
        Put the device into a state in which it is ready to release the labware at the specified handover position.

        Args:
          HandoverPosition: Indicates the position where the labware will
            be handed over.
          InternalPosition: Indicates the position which the labware will
            be retrieved from within the device, e.g. internal storage
            positions of an incubator.

        Raises:
          InvalidCommandSequence: The issued command does not follow the
            sequence of commands for the device according to its role in
            the labware transfer.
        """

    @abc.abstractmethod
    @sila.ObservableCommand()
    async def put_labware(
        self,
        handover_position: HandoverPosition,
        intermediate_actions: list[str],
        *,
        status: sila.Status,
    ) -> None:
        """
        Place the currently processed labware item at the specified handover position.

        Handover position is sent to the active source device after a
        "Prepare For Output" command.

        Args:
          HandoverPosition: Indicates the position the labware item will
            be moved to.
          IntermediateActions: Specifies one or more commands that have
            to be executed within the command sequence (e.g. removing a
            lid).
            The order of execution is specified by order within the given
            list.
            Each entry must be one of the commands returned by the
            AvailableIntermediateCommandExecutions property.

        Raises:
          InvalidCommandSequence: The issued command does not follow the
            sequence of commands for the device according to its role in
            the labware transfer.
          LabwareNotPlaced: Placing the labware item at the destination
            device failed.
        """

    @abc.abstractmethod
    @sila.ObservableCommand()
    async def get_labware(
        self,
        handover_position: HandoverPosition,
        intermediate_actions: list[str],
        *,
        status: sila.Status,
    ) -> None:
        """
        Retrieve a labware item from the specified handover position.

        Handover position is sent to the active destination device after
        a "Prepare For Input" command.

        Args:
          HandoverPosition: Indicates the position where the labware will
            be retrieved from.
          IntermediateActions: Specifies one or more commands that have
            to be executed within the command sequence (e.g. removing a
            lid).
            The order of execution is specified by order within the given
            list.
            Each entry must be one of the commands returned by the
            AvailableIntermediateCommandExecutions property.

        Raises:
          InvalidCommandSequence: The issued command does not follow the
            sequence of commands for the device according to its role in
            the labware transfer.
          LabwareNotPicked: Picking up the labware item from the source
            device failed.
        """



import logging
import pathlib

import click

from unitelabs.cdk import utils
from unitelabs.cdk.config import read_config_file
from unitelabs.cdk.logging import configure_logging

from ..main import run


class TLSConfigurationError(Exception):
    """TLS Configuration is invalid."""


@click.command()
@click.option(
    "--app",
    type=str,
    metavar="IMPORT",
    default=utils.find_factory,
    show_default=False,
    help="The application factory function to load, in the form 'module:name'.",
    envvar="UNITELABS_CDK_APP",
)
@click.option(
    "-cfg",
    "--config-path",
    type=click.Path(path_type=pathlib.Path),
    help="Path to the configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help=(
        "Increase the verbosity of the default logger. "
        "Use a custom log-config for fine grained handling."
        "When used together with a log-config, this will override the root logger's level."
    ),
)
@utils.coroutine
async def start(app, config_path: pathlib.Path, verbose: int) -> None:  # noqa: ANN001
    """Application Entrypoint."""

    config = read_config_file(config_path)

    log_level = logging.ERROR - verbose * 10 if verbose else None
    log_config = config.get("logging", None)
    configure_logging(log_config, log_level=log_level)

    await run(app, config=config)



from sila import Element, Native


def to_sila(value: Native, responses: dict[str, Element]) -> Native:
    """
    Convert a (intermediate) command response to a SiLA native value.

    Args:
      value: The command responses value to convert.
      responses: The SiLA data type of the message.

    Returns:
      The converted SiLA native value.
    """

    if isinstance(value, dict) and responses.keys() != value.keys():
        value = (value,)

    if not isinstance(value, (tuple, dict)):
        if keys := responses.keys():
            identifier = next(iter(keys))
            value = {identifier: value}
        else:
            value = {}

    if isinstance(value, tuple):
        values = {}
        for index, name in enumerate(responses.keys()):
            if index >= len(value):
                msg = f"Expected {len(responses)} elements in tuple, received {value}."
                raise ValueError(msg)
            values[name] = value[index]

        value = values

    return value



# ruff: noqa: D205, D401, E501

import dataclasses

from unitelabs.cdk import sila


@dataclasses.dataclass
class TestStructure(sila.CustomDataType):
    """
    An example Structure data type containing all SiLA basic types.

    Attributes:
      StringTypeValue: A value of SiLA data type String.
      IntegerTypeValue: A value of SiLA data type Integer.
      RealTypeValue: A value of SiLA data type Real.
      BooleanTypeValue: A value of SiLA data type Boolean.
      BinaryTypeValue: A value of SiLA data type Binary.
      DateTypeValue: A value of SiLA data type Date.
      TimeTypeValue: A value of SiLA data type Time.
      TimestampTypeValue: A value of SiLA data type Timestamp
      AnyTypeValue: A value of SiLA data type Any.
    """

    string_type_value: str
    integer_type_value: int
    real_type_value: float
    boolean_type_value: bool
    binary_type_value: bytes
    date_type_value: sila.datetime.date
    time_type_value: sila.datetime.time
    timestamp_type_value: sila.datetime.datetime
    any_type_value: sila.Any


@dataclasses.dataclass
class InnerStructure:
    """
    A structure type that is part of the middle structure.

    Attributes:
      InnerStringTypeValue: A value of SiLA data type String contained in the innermost structure.
      InnerIntegerTypeValue: A value of SiLA data type Integer contained in the innermost structure.
    """

    inner_string_type_value: str
    inner_integer_type_value: int


@dataclasses.dataclass
class MiddleStructure:
    """
    Another structure type that is part of the outer structure.

    Attributes:
      MiddleStringTypeValue: A value of SiLA data type String contained in the middle structure.
      MiddleIntegerTypeValue: A value of SiLA data type Integer contained in the middle structure.
      InnerStructure: A structure type that is part of the middle structure.
    """

    middle_string_type_value: str
    middle_integer_type_value: int
    inner_structure: InnerStructure


@dataclasses.dataclass
class DeepStructure(sila.CustomDataType):
    """
    An example Structure data type that contains other structures within.

    Attributes:
      OuterStringTypeValue: A value of SiLA data type String contained in the topmost structure.
      OuterIntegerTypeValue: A value of SiLA data type Integer contained in the topmost structure.
      MiddleStructure: Another structure type that is part of the outer structure.
    """

    outer_string_type_value: str
    outer_integer_type_value: int
    middle_structure: MiddleStructure


class StructureDataTypeTest(sila.Feature):
    """Provides commands and properties to set or respectively get SiLA Structure Data Type values via command parameters or property responses respectively."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    # Simple Structure type

    @sila.UnobservableCommand()
    def echo_structure_value(self, structure_value: TestStructure) -> TestStructure:
        """
        Receives a structure value and returns the structure that has been received (binary value is expected to be an embedded value, any typer value is expected to be a Basic type).

        Args:
          StructureValue: The Structure value to be returned.

        Returns:
          ReceivedValues: The structure that has been received.
        """

        return structure_value

    @sila.UnobservableProperty()
    def structure_value(self) -> TestStructure:
        """
        Returns a structure with the following elements values:
        - String value = 'SiLA2_Test_String_Value'
        - Integer value = 5124
        - Real value = 3.1415926
        - Boolean value = true
        - Binary value = embedded string 'SiLA2_Binary_String_Value'
        - Date value = 05.08.2022 respective 08/05/2022
        - Time value = 12:34:56.789
        - Timestamp value = 2022-08-05 12:34:56.789
        - Any type value = string 'SiLA2_Any_Type_String_Value'.
        """

        return TestStructure(
            string_type_value="SiLA2_Test_String_Value",
            integer_type_value=5124,
            real_type_value=3.1415926,
            boolean_type_value=True,
            binary_type_value=b"SiLA2_Binary_String_Value",
            date_type_value=sila.datetime.date(
                year=2022, month=8, day=5, tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2))
            ),
            time_type_value=sila.datetime.time(
                hour=12,
                minute=34,
                second=56,
                microsecond=789000,
                tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
            ),
            timestamp_type_value=sila.datetime.datetime(
                year=2022,
                month=8,
                day=5,
                hour=12,
                minute=34,
                second=56,
                microsecond=789000,
                tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
            ),
            any_type_value="SiLA2_Any_Type_String_Value",
        )

    # Deep Structure type

    @sila.UnobservableCommand()
    def echo_deep_structure_value(self, deep_structure_value: DeepStructure) -> DeepStructure:
        """
        Receives a multilevel structure value and returns the structure that has been received.

        Args:
          DeepStructureValue: The deep Structure value to be set.

        Returns:
          ReceivedValues: The structure that has been received.
        """

        return deep_structure_value

    @sila.UnobservableProperty()
    def deep_structure_value(self) -> DeepStructure:
        """
        Returns a multilevel structure with the following values:
        - string value = 'Outer_Test_String'
        - integer value = 1111
        - middle structure value =
          - string value = 'Middle_Test_String'
          - integer value = 2222
          - inner structure value =
            - string value = 'Inner_Test_String'
            - integer value = 3333.
        """

        return DeepStructure(
            outer_string_type_value="Outer_Test_String",
            outer_integer_type_value=1111,
            middle_structure=MiddleStructure(
                middle_string_type_value="Middle_Test_String",
                middle_integer_type_value=2222,
                inner_structure=InnerStructure(
                    inner_string_type_value="Inner_Test_String",
                    inner_integer_type_value=3333,
                ),
            ),
        )



from sila.framework.errors.defined_execution_error import DefinedExecutionError
from unitelabs.cdk.sila.common.errors import define_error


class TestDefineError:
    # Defines an error with exception type.
    def test_define_error_with_type(self):
        exception = Exception

        defined_execution_error = define_error(exception)

        assert issubclass(defined_execution_error, DefinedExecutionError)
        assert defined_execution_error.identifier == "Exception"
        assert defined_execution_error.display_name == "Exception"
        assert defined_execution_error.description == "Common base class for all non-exit exceptions."

    # Defines an error with exception instance.
    def test_define_error_with_instance(self):
        exception = RuntimeError()

        defined_execution_error = define_error(exception)

        assert issubclass(defined_execution_error, DefinedExecutionError)
        assert defined_execution_error.identifier == "RuntimeError"
        assert defined_execution_error.display_name == "Runtime Error"
        assert defined_execution_error.description == "Unspecified run-time error."

    # Defines an error with exception instance.
    def test_define_error_with_message(self):
        exception = ValueError("Hello, World!")

        defined_execution_error = define_error(exception)

        assert issubclass(defined_execution_error, DefinedExecutionError)
        assert defined_execution_error.identifier == "ValueError"
        assert defined_execution_error.display_name == "Value Error"
        assert defined_execution_error.description == "Inappropriate argument value (of correct type)."



import dataclasses

import typing_extensions as typing

import sila

T = typing.TypeVar("T", bound=dict[str, sila.Native])


@dataclasses.dataclass
class Structure(typing.Generic[T], sila.Structure[T]):
    """Structure that converts from and to a python class instead of a dictionary."""

    _class: typing.ClassVar[type] = dataclasses.field(init=None)

    @typing.override
    @classmethod
    async def from_native(
        cls,
        context: "sila.Context",
        value: T | None = None,
        /,
        *,
        execution: typing.Optional["sila.Execution"] = None,
    ) -> typing.Self:
        if not issubclass(cls._class, dict) and isinstance(value, cls._class):
            value = {key: getattr(value, key, None) for key in cls.elements}

        return await super().from_native(context, value, execution=execution)

    @typing.override
    async def to_native(self, context: "sila.Context", /) -> T:
        values = await super().to_native(context)

        return self._class(**values)



import unittest.mock

from unitelabs.cdk.sila.common.decorator import Decorator


class TestCall:
    def test_should_call_decorator(self):
        # Create decorator
        decorator = Decorator(
            identifier="Identifier",
            name="Name",
            errors=[Exception],
            enabled=True,
        )

        # Call decorator
        function = unittest.mock.Mock()
        decorator(function)

        # Assert that the method returns the correct value
        assert decorator._function == function
        assert getattr(function, "__handler") == decorator


class TestCopy:
    def test_should_copy_decorator(self):
        # Create decorator
        decorator = Decorator(
            identifier="Identifier",
            name="Name",
            errors=[Exception],
            enabled=True,
        )

        # Copy decorator
        copy = decorator.clone()

        # Assert that the method returns the correct value
        assert copy is not decorator
        assert copy._identifier == "Identifier"
        assert copy._name == "Name"
        assert copy._description == ""
        assert copy._enabled is True
        assert copy._errors == [Exception]



import dataclasses

import typing_extensions as typing

import sila


@dataclasses.dataclass
class Any(sila.Any):
    """Any data type that converts to its native python type."""

    @typing.override
    async def to_native(self, context: "sila.Context", /) -> sila.Native:
        return await self.value.to_native(context)



import enum


class Default(enum.Enum):
    """A sentinel value used to indicate that a `Subject`, `Publisher`, or `Subscription` has not been updated yet."""

    token = 0


_DEFAULT_VALUE = Default.token



import asyncio
import datetime

import typing_extensions as typing

from unitelabs.cdk import sila


class MultiClientTest(sila.Feature):
    """Test different server behaviors when multiple clients request execution of the same command."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.ObservableCommand(mode=sila.ExecutionMode.PARALLEL)
    async def run_in_parallel(
        self,
        duration: typing.Annotated[
            float, sila.constraints.Unit(label="s", components=[sila.constraints.Unit.Component("Second")])
        ],
        *,
        status: sila.Status,
    ) -> None:
        """
        Multiple invocations of this command will be running in parallel.

        Args:
          Duration: The duration of the command execution
        """

        status.update(progress=0, remaining_time=datetime.timedelta(seconds=duration))

        await asyncio.sleep(duration)

    @sila.ObservableCommand(mode=sila.ExecutionMode.QUEUED)
    async def run_queued(
        self,
        duration: typing.Annotated[
            float, sila.constraints.Unit(label="s", components=[sila.constraints.Unit.Component("Second")])
        ],
        *,
        status: sila.Status,
    ) -> None:
        """
        Multiple invocations of this command will be queued.

        Args:
          Duration: The duration of the command execution
        """

        status.update(progress=0, remaining_time=datetime.timedelta(seconds=duration))

        await asyncio.sleep(duration)

    @sila.ObservableCommand(mode=sila.ExecutionMode.QUEUED)
    async def reject_parallel_execution(
        self,
        duration: typing.Annotated[
            float, sila.constraints.Unit(label="s", components=[sila.constraints.Unit.Component("Second")])
        ],
        *,
        status: sila.Status,
    ) -> None:
        """
        Invocations will be rejected, if there is another command instance already running.

        Args:
          Duration: The duration of the command execution
        """

        status.update(progress=0, remaining_time=datetime.timedelta(seconds=duration))

        await asyncio.sleep(duration)



import asyncio
import collections.abc
import functools
import importlib
import inspect
import signal
import sys
import types

import typing_extensions as typing

from .connector import Connector

T = typing.TypeVar("T")


Generator: typing.TypeAlias = collections.abc.Generator[T, None, None] | collections.abc.AsyncGenerator[T, None]


AppFactory = collections.abc.Callable[
    ...,
    Connector | collections.abc.Awaitable[Connector] | Generator,
]


async def run(app: str | AppFactory | Connector, /, config: dict | None = None) -> None:
    """
    Run the given application.

    Args:
      app: Either an entrypoint reference (e.g. `unitelabs.awesome_instrument:create_app`),
        an app factory method or the app instance directly.
      config: Optional configuration to pass to the app factory.
    """

    if isinstance(app, str):
        app = await load(app)

    if callable(app):
        if "config" in inspect.signature(app).parameters:
            from unitelabs.cdk.config import get_connector_config

            configuration = get_connector_config().validate(config)
            app = functools.partial(app, config=configuration)

        app = await init(app_factory=app)

    for signum in (signal.SIGTERM, signal.SIGINT):
        if sys.platform == "win32":
            signal.signal(signum, functools.partial(signal_handler, app=app))
        else:
            asyncio.get_running_loop().add_signal_handler(
                signum, functools.partial(signal_handler, signum, None, app=app)
            )

    try:
        start = asyncio.create_task(app.start())
        termination = asyncio.create_task(app.wait_for_termination())

        _, pending = await asyncio.wait(
            [start, termination],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if start in pending:
            start.cancel()

        if termination in pending:
            await termination
    finally:
        await app.stop()


async def load(entrypoint: str) -> AppFactory | Connector:
    """
    Dynamically import an app instance or factory from the given entrypoint.

    Args:
      entrypoint: Where to find the app factory formatted as "module:name",
        (e.g. `unitelabs.awesome_instrument:create_app`)

    Returns:
      The app instance or factory at the given entrypoint.
    """

    module_name, _, factory_name = entrypoint.partition(":")

    if not module_name or not factory_name:
        msg = f"Entrypoint '{entrypoint}' must be in format '<module>:<attribute>'."
        raise ValueError(msg)

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        msg = f"Could not import module '{module_name}': Module not found."
        raise ValueError(msg) from None

    instance = module
    try:
        for attr in factory_name.split("."):
            instance = getattr(instance, attr)
    except AttributeError:
        msg = f"Attribute '{factory_name}' not found in module '{module_name}'."
        raise ValueError(msg) from None

    if callable(instance):
        return typing.cast(AppFactory, instance)

    if isinstance(instance, Connector):
        return instance

    msg = f"Attribute '{factory_name}' in module '{module_name}' is must be app instance or factory."
    raise ValueError(msg)


def signal_handler(signum: int, frame: types.FrameType | None, app: Connector) -> None:  # noqa: ARG001
    """
    Gracefully handle received signals.

    Args:
      signum: The received signal number.
      frame: The current stack frame.
      app: The running connector application.
    """

    app.logger.info("Received signal %s. Shutting down...", signal.Signals(signum).name)
    if sys.platform == "win32":
        signal.signal(signum, signal.SIG_DFL)
    else:
        loop = asyncio.get_running_loop()
        loop.remove_signal_handler(signum)
    app._shutdown.set()


async def init(app_factory: AppFactory) -> Connector:
    """
    Use the provided factory method to init a new `Connector`.

    Args:
      app_factory: The factory method to call.

    Returns:
      The initialized `Connector` which shutdown handlers attached.

    Raises:
      ValueError: If `app_factory` does not follow the required
        interface.
    """

    app = None

    if inspect.isasyncgenfunction(app_factory) or inspect.isgeneratorfunction(app_factory):
        generator: Generator[Connector] = app_factory()
        generator = _sync_to_async_gen(generator)

        try:
            app = await generator.__anext__()
        except (StopAsyncIteration, StopIteration):
            msg = "Unable to create app: `create_app` did not yield a value."
            raise ValueError(msg) from None

        shutdown_handler = functools.partial(_shutdown_yield, generator)
        app.on_shutdown(handler=shutdown_handler)

    elif inspect.iscoroutinefunction(app_factory):
        app = await app_factory()

    elif inspect.isfunction(app_factory):
        app = app_factory()

    if app is None:
        msg = f"Invalid `create_app`: '{app_factory}'. Provide a callable function that returns a Connector."
        raise ValueError(msg)

    if not isinstance(app, Connector):
        msg = f"Expected app to be of type 'Connector', received '{app.__class__.__name__}'."
        raise ValueError(msg) from None

    return app


async def _shutdown_yield(generator: collections.abc.AsyncGenerator[T, None]) -> None:
    """
    Execute the shutdown of a factory function.

    Achieved by advancing the iterator after the yield to
    ensure the iteration ends (if not it means there is
    more than one yield in the function).

    Args:
      generator: The factory function to create the app.
    """

    try:
        await generator.__anext__()
    except (StopAsyncIteration, StopIteration):
        pass
    else:
        await _shutdown_yield(generator)


async def _sync_to_async_gen(generator: Generator[T]) -> collections.abc.AsyncGenerator[T, None]:
    """
    Wrap any generator into an async generator.

    Args:
      generator: The generator to wrap as async.

    Returns:
      The async generator.
    """

    if inspect.isasyncgen(generator):
        async for item in generator:
            yield item

        return

    if inspect.isgenerator(generator):
        while True:
            try:
                yield next(generator)
            except StopIteration:
                return



import datetime
import weakref


from sila.server import CommandExecution


class Status:
    """A class representing the status of an observablØe command execution."""

    def __init__(self, command_execution: CommandExecution):
        self.command_execution: CommandExecution = weakref.proxy(command_execution)

    def update(
        self,
        progress: float | None = None,
        remaining_time: datetime.timedelta | None = None,
        updated_lifetime: datetime.timedelta | None = None,
    ) -> None:
        """Update the execution status of an observable command execution."""
        self.command_execution.update_execution_info(
            progress=progress,
            remaining_time=remaining_time,
            updated_lifetime=updated_lifetime,
        )



import asyncio
import signal
import unittest.mock

import pytest

from unitelabs.cdk.connector import Connector
from unitelabs.cdk.main import init, load, run, signal_handler


@pytest.fixture
async def app():
    return Connector()


class TestRun:
    async def test_should_run_with_factory(self, app: Connector):
        app.start = unittest.mock.AsyncMock()
        app.wait_for_termination = unittest.mock.AsyncMock()
        app.stop = unittest.mock.AsyncMock()

        async def app_factory() -> Connector:
            return app

        # Run
        await run(app_factory)

        # Assert that the method returns the correct value
        app.start.assert_called_once_with()
        app.stop.assert_called_once_with()

    async def test_should_run_with_instance(self, app: Connector):
        app.start = unittest.mock.AsyncMock()
        app.wait_for_termination = unittest.mock.AsyncMock()
        app.stop = unittest.mock.AsyncMock()

        # Run
        await run(app)

        # Assert that the method returns the correct value
        app.start.assert_called_once_with()
        app.stop.assert_called_once_with()


class TestSignalHandler:
    async def test_signal_handler_exists_with_signum(self):
        app = unittest.mock.Mock(_shutdown=asyncio.Event())

        signal_handler(signum=signal.SIGINT, frame=None, app=app)

        assert app._shutdown.is_set()


class TestLoad:
    async def test_should_load_factory(self):
        # Load
        result = await load("unitelabs.cdk.main:load")

        # Assert that the method returns the correct value
        assert result == load

    async def test_should_raise_on_invalid_format(self):
        # Load
        with pytest.raises(ValueError, match=r"Entrypoint 'unitelabs' must be in format '<module>:<attribute>'\."):
            await load("unitelabs")

    async def test_should_raise_on_unknown_module(self):
        # Load
        with pytest.raises(ValueError, match=r"Could not import module 'module': Module not found\."):
            await load("module:attribute")

    async def test_should_raise_on_unknown_attribute(self):
        # Load
        with pytest.raises(ValueError, match=r"Attribute 'attribute' not found in module 'unitelabs.cdk.main'\."):
            await load("unitelabs.cdk.main:attribute")

    async def test_should_raise_on_non_connector(self):
        # Load
        with pytest.raises(
            ValueError, match=r"Attribute 'T' in module 'unitelabs.cdk.main' is must be app instance or factory\."
        ):
            await load("unitelabs.cdk.main:T")


class TestInit:
    async def test_should_init_function(self, app: Connector):
        # Create factory
        def factory():
            return app

        # Init
        result = await init(factory)

        # Assert that the method returns the correct value
        assert result == app
        assert not result._shutdown_handlers

    async def test_should_init_async_function(self, app: Connector):
        # Create factory
        async def factory():
            return app

        # Init
        result = await init(factory)

        # Assert that the method returns the correct value
        assert result == app
        assert not result._shutdown_handlers

    async def test_should_init_generator(self, app: Connector):
        # Create factory
        def factory():
            yield app

        # Init
        result = await init(factory)

        # Assert that the method returns the correct value
        assert result == app

    async def test_should_init_async_generator(self, app: Connector):
        # Create factory
        async def factory():
            yield app

        # Init
        result = await init(factory)

        # Assert that the method returns the correct value
        assert result == app

    async def test_should_register_shutdown_handler(self, app: Connector):
        # Create factory
        callback = unittest.mock.Mock()

        async def factory():
            yield app
            callback()

        # Init
        result = await init(factory)
        await result.stop()

        # Assert that the method returns the correct value
        callback.assert_called_once_with()

    async def test_should_register_async_shutdown_handler(self, app: Connector):
        # Create factory
        callback = unittest.mock.Mock()

        async def factory():
            yield app
            callback()

        # Init
        result = await init(factory)
        await result.stop()

        # Assert that the method returns the correct value
        callback.assert_called_once_with()

    async def test_should_register_terminate_on_multiple_yields(self, app: Connector):
        # Create factory
        callback = unittest.mock.Mock()

        async def factory():
            yield app
            yield
            callback()

        # Init
        result = await init(factory)
        await result.stop()

        # Assert that the method returns the correct value
        callback.assert_called_once_with()

    async def test_should_raise_on_empty_generator(self):
        # Create factory
        def factory():
            return
            yield

        # Init
        with pytest.raises(ValueError, match=r"Unable to create app: `create_app` did not yield a value\."):
            await init(factory)

    async def test_should_raise_on_non_connector(self):
        # Create factory
        def factory():
            return object()

        # Init
        with pytest.raises(ValueError, match=r"Expected app to be of type 'Connector', received 'object'\."):
            await init(factory)  # type: ignore



import collections.abc
import dataclasses
import inspect
import re
import warnings

import griffe
import typing_extensions as typing

from sila import DefinedExecutionError, Element, Feature

from .name import to_display_name, to_identifier


@dataclasses.dataclass
class Docstring:
    """A parsed docstring of a function or class."""

    description: str = ""
    parameters: dict[str, Element] = dataclasses.field(default_factory=dict)
    yields: dict[str, Element] = dataclasses.field(default_factory=dict)
    returns: dict[str, Element] = dataclasses.field(default_factory=dict)
    raises: dict[str, DefinedExecutionError] = dataclasses.field(default_factory=dict)


def parse_docstring(functionOrClass: collections.abc.Callable | type, /, feature: Feature) -> Docstring:
    """
    Parse the docstring of a function or class.

    Args:
      functionOrClass: The function or class to parse the docstring of.
      feature: The parent feature of the function or class.

    Returns:
      The parsed docstring.
    """

    from ..command import Intermediate, Status
    from ..metadata import Metadata

    # Types
    parameters: list[griffe.DocstringParameter] = []
    yields: list[griffe.DocstringYield] = []
    returns: list[griffe.DocstringReturn] = []

    if dataclasses.is_dataclass(functionOrClass):
        parameters = [
            griffe.DocstringParameter(name=field.name, description="", annotation=field.type, value=field.name)
            for field in dataclasses.fields(functionOrClass)
        ]
    elif isinstance(functionOrClass, type):
        parameters = [
            griffe.DocstringParameter(name=name, description="", annotation=annotation, value=name)
            for name, annotation in typing.get_type_hints(functionOrClass).items()
        ]
    else:
        signature = inspect.signature(functionOrClass)
        returns = [
            griffe.DocstringReturn(name=f"response_{i}", description="", annotation=return_annotation)
            for i, return_annotation in enumerate(get_types(signature.return_annotation))
        ]

        for i, (name, item) in enumerate(signature.parameters.items()):
            origin = typing.get_origin(item.annotation) or item.annotation

            if origin is Intermediate:
                args = typing.get_args(item.annotation)
                yields = [
                    griffe.DocstringYield(
                        name=f"intermediate_response_{j}", description="", annotation=yield_annotation
                    )
                    for j, yield_annotation in enumerate(get_types(args[0] if args else inspect._empty))
                ]
                continue

            if origin is typing.Annotated and typing.get_args(item.annotation)[0] is Metadata:
                continue

            if item.annotation is Status:
                continue

            if i == 0 and name in ("self", "cls"):
                continue

            parameters.append(
                griffe.DocstringParameter(name=item.name, description="", annotation=item.annotation, value=name)
            )

    # Docstring
    doc = inspect.getdoc(functionOrClass) or ""
    parent = griffe.Function(name=functionOrClass.__name__)
    docstring = griffe.Docstring(doc, parent=parent)

    if ".. parameter::" in doc or ".. yield::" in doc or ".. return::" in doc:
        _warn(
            "Using reStructuredText based docstrings is deprecated. Please use google based formatting instead.",
            functionOrClass,
            category=DeprecationWarning,
        )
        docstring.parent.parameters = griffe.Parameters(*parameters)
        sections = parse_restructured(docstring)
    else:
        sections = griffe.parse(
            docstring, "google", returns_multiple_items=True, returns_named_value=True, warnings=False
        )

    result = Docstring()

    for section in sections:
        if isinstance(section, griffe.DocstringSectionText):
            result.description += section.value

        if isinstance(section, griffe.DocstringSectionExamples):
            result.description += "\n\n"
            for example in section.value:
                result.description += example[1] + "\n"

        if isinstance(section, griffe.DocstringSectionParameters | griffe.DocstringSectionAttributes):
            for i, item in enumerate(section.value):
                if len(parameters) <= i:
                    _warn(
                        f"Found documentation '{item.name}: {item.description}' for non-existent parameter.",
                        functionOrClass,
                    )
                    continue

                if item.annotation is not None:
                    _warn(
                        "Using type annotations inside your docstring is not allowed and will be ignored.",
                        functionOrClass,
                    )

                parameters[i].name = item.name or parameters[i].name
                parameters[i].description = item.description

        if isinstance(section, griffe.DocstringSectionYields):
            for i, item in enumerate(section.value):
                if len(yields) <= i:
                    _warn(
                        f"Found documentation '{item.name}: {item.description}' for non-existent yield.",
                        functionOrClass,
                    )
                    continue

                if item.annotation is not None:
                    _warn(
                        "Using type annotations inside your docstring is not allowed and will be ignored.",
                        functionOrClass,
                    )

                yields[i].name = item.name or yields[i].name
                yields[i].description = item.description

        if isinstance(section, griffe.DocstringSectionReturns):
            for i, item in enumerate(section.value):
                if len(returns) <= i:
                    _warn(
                        f"Found documentation '{item.name}: {item.description}' for non-existent return.",
                        functionOrClass,
                    )
                    continue

                if item.annotation is not None:
                    _warn(
                        "Using type annotations inside your docstring is not allowed and will be ignored.",
                        functionOrClass,
                    )

                returns[i].name = item.name or returns[i].name
                returns[i].description = item.description

        if isinstance(section, griffe.DocstringSectionRaises):
            for item in section.value:
                name = to_display_name(str(item.annotation))
                key = name.replace(" ", "_").lower()

                result.raises[key] = DefinedExecutionError.create(
                    identifier=to_identifier(name), display_name=name, description=item.description
                )

    result.parameters = dict(create_element(item, feature) for item in parameters)
    result.yields = dict(create_element(item, feature) for item in yields)
    result.returns = dict(create_element(item, feature) for item in returns)

    return result


def parse_restructured(docstring: griffe.Docstring) -> list[griffe.DocstringSection]:
    """
    Parse documentation strings in reStructuredText format.

    Args:
      docstring: The docstring to parse.

    Returns:
      The parsed docstring.
    """

    sections: list[griffe.DocstringSection] = []
    directives = re.split(r"^\.\. *([^:]+):: *", docstring.value, flags=re.MULTILINE)

    if not isinstance(docstring.parent, griffe.Function):
        return sections

    section = griffe.DocstringSectionText(value=inspect.cleandoc(directives.pop(0)))
    parameters = griffe.DocstringSectionParameters(value=[])
    yields = griffe.DocstringSectionYields(value=[])
    returns = griffe.DocstringSectionReturns(value=[])

    for i in range(0, len(directives), 2):
        key = directives[i]
        params = re.split(r"^ *:([^:]+): *", directives[i + 1], flags=re.MULTILINE)
        name = ""
        description = inspect.cleandoc(params.pop(0)).replace("\n", " ")

        par = {params[i]: params[i + 1] for i in range(0, len(params), 2)}
        for param_k, param_v in par.items():
            if param_k == "name":
                name = inspect.cleandoc(param_v).replace("\n", " ")

        if key == "parameter":
            parameter = docstring.parent.parameters[len(parameters.value)]

            parameters.value.append(
                griffe.DocstringParameter(
                    name=name or to_display_name(parameter.name),
                    description=description,
                )
            )

        if key == "yield":
            yields.value.append(griffe.DocstringYield(name=name.replace(" ", "_"), description=description))

        if key == "return":
            returns.value.append(griffe.DocstringReturn(name=name.replace(" ", "_"), description=description))

    sections.append(section)

    if parameters.value:
        sections.append(parameters)

    if yields.value:
        sections.append(yields)

    if returns.value:
        sections.append(returns)

    return sections


def create_element(
    item: griffe.DocstringParameter | griffe.DocstringYield | griffe.DocstringReturn,
    feature: Feature,
    default: str = "",
) -> tuple[str, Element]:
    """
    Create an element from a docstring item.

    Args:
      item: The docstring item to create an element from.
      feature: The parent feature of the element.
      default: The default name of the element.

    Returns:
      The created element.
    """

    from ..data_types import infer

    name = to_display_name(item.name or default)
    key = item.value or name.replace(" ", "_").lower()

    return key, Element(
        identifier=to_identifier(name),
        display_name=name,
        description=item.description,
        data_type=infer(item.annotation, feature),
    )


def _warn(msg: str, functionOrClass: collections.abc.Callable | type, category: type[Warning] = UserWarning) -> None:
    warnings.warn_explicit(
        msg,
        category=category,
        filename=inspect.getfile(inspect.unwrap(functionOrClass)),
        lineno=inspect.getsourcelines(inspect.unwrap(functionOrClass))[1],
    )


def get_types(annotation: type) -> list[type]:
    """
    Get the types from an annotation.

    Args:
      annotation: The annotation to get the types from.

    Returns:
      The types.
    """

    return (
        list(typing.get_args(annotation))
        if (typing.get_origin(annotation) or annotation) is tuple
        else [annotation]
        if annotation is not None and annotation is not type(None)
        else []
    )



import dataclasses
import datetime

from sila.framework.constraints import MaximalExclusive as SiLAMaximalExclusive
from sila.framework.data_types import Date, Integer, Real, Time, Timestamp


@dataclasses.dataclass
class MaximalExclusive(SiLAMaximalExclusive):
    """A constraint that enforces an upper exclusive bound on a value."""

    value: int | float | datetime.date | datetime.time | datetime.datetime

    def __post_init__(self):
        if isinstance(self.value, int):
            self.value = Integer(self.value)
        elif isinstance(self.value, float):
            self.value = Real(self.value)
        elif isinstance(self.value, datetime.datetime):
            self.value = Timestamp.from_datetime(self.value)
        elif isinstance(self.value, datetime.date):
            self.value = Date.from_date(self.value)
        elif isinstance(self.value, datetime.time):
            self.value = Time.from_time(self.value)

        else:
            msg = (
                "Expected type of str, int, float, datetime.date, datetime.time, datetime.datetime, "
                f"Instead received type '{type(self.value)}'."
            )
            raise TypeError(msg)



import datetime
import unittest.mock
import zoneinfo

import pytest

from unitelabs.cdk import sila
from unitelabs.cdk.sila.constraints import Set


class TestValidate:
    @pytest.mark.parametrize(
        "value,validation,error",
        [
            pytest.param(["a", "b"], sila.data_types.String("a"), False, id="valid-str"),
            pytest.param(["a", "b"], sila.data_types.String("c"), True, id="invalid-str"),
            pytest.param([1, 2], sila.data_types.Integer(1), False, id="valid-int"),
            pytest.param([1, 2], sila.data_types.Integer(3), True, id="invalid-int"),
            pytest.param([1.0, 1.1], sila.data_types.Real(1.0), False, id="valid-float"),
            pytest.param([1.0, 1.1], sila.data_types.Real(1.2), True, id="invalid-float"),
            pytest.param(
                [datetime.date(2023, 1, 1), datetime.date(2024, 1, 1)],
                sila.data_types.Date(2023, 1, 1),
                False,
                id="valid-date",
            ),
            pytest.param(
                [datetime.date(2023, 1, 1), datetime.date(2024, 1, 1)],
                sila.data_types.Date(2023, 1, 2),
                True,
                id="invalid-date",
            ),
            pytest.param(
                [
                    datetime.time(12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                    datetime.time(12, 1, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                ],
                sila.data_types.Time(12, 0, 0, 0),
                False,
                id="valid-time",
            ),
            pytest.param(
                [
                    datetime.time(12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                    datetime.time(12, 1, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                ],
                sila.data_types.Time(12, 0, 0, 1),
                True,
                id="invalid-time",
            ),
            pytest.param(
                [
                    datetime.datetime(2023, 1, 1, 12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                    datetime.datetime(2023, 1, 1, 13, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                ],
                sila.data_types.Timestamp(2023, 1, 1, 12, 0, 0, 0),
                False,
                id="valid-timestamp",
            ),
            pytest.param(
                [
                    datetime.datetime(2023, 1, 1, 12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                    datetime.datetime(2023, 1, 1, 13, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                ],
                sila.data_types.Timestamp(2023, 1, 1, 14, 0, 0, 0),
                True,
                id="invalid-timestamp",
            ),
        ],
    )
    async def test_validation(self, value, validation, error):
        constraint = Set(value)
        if not error:
            assert await constraint.validate(validation)
        else:
            with pytest.raises(ValueError):
                await constraint.validate(validation)

    async def test_should_raise_type_error_unsupported_type(self):
        with pytest.raises(TypeError):
            Set(unittest.mock.Mock())



import weakref

import typing_extensions as typing
from sila.server import CommandExecution, Element

from ..data_types import to_sila

T = typing.TypeVar("T")


class Intermediate(typing.Generic[T]):
    """A class representing an intermediate response in a command execution."""

    def __init__(self, command_execution: CommandExecution, responses: dict[str, Element]):
        self.command_execution: CommandExecution = weakref.proxy(command_execution)
        self.responses = responses

    def send(self, *responses: T) -> None:
        """Send an intermediate response."""

        self.command_execution.send_intermediate_responses(to_sila(responses, self.responses))



from importlib.metadata import version

from .config import (
    CloudServerConfig,
    ConfigurationError,
    ConnectorBaseConfig,
    Field,
    SiLAServerConfig,
    UnsupportedConfigFiletype,
    validate_config,
    validate_field,
)
from .connector import Connector
from .logging import create_logger
from .main import AppFactory, run
from .subscriptions import Publisher, Subject, Subscription

__version__ = version("unitelabs-cdk")
__all__ = [
    "AppFactory",
    "CloudServerConfig",
    "ConfigurationError",
    "Connector",
    "ConnectorBaseConfig",
    "Field",
    "Publisher",
    "SiLAServerConfig",
    "Subject",
    "Subscription",
    "UnsupportedConfigFiletype",
    "__version__",
    "create_logger",
    "run",
    "validate_config",
    "validate_field",
]



import collections.abc
import dataclasses
import datetime
import enum
import inspect
import types

import typing_extensions as typing
from sila.server import (
    Binary,
    Boolean,
    Constrained,
    DataType,
    Date,
    Integer,
    List,
    Native,
    Real,
    String,
    Time,
    Timestamp,
    Void,
)

from ..utils import parse_docstring
from .any import Any
from .custom_data_type import CustomDataType
from .enum import Enum
from .structure import Structure

if typing.TYPE_CHECKING:
    from ..common import Feature


def infer(annotation: type, feature: "Feature") -> type[DataType]:
    """
    Infer the SiLA data type from a given python type annotation.

    Args:
      annotation: The python type annotation.
      feature: The feature.

    Returns:
      The corresponding SiLA data type.
    """

    if annotation == inspect._empty:
        return Any

    origin = typing.get_origin(annotation) or annotation

    if origin is None:
        return Void
    if origin is typing.Annotated:
        args = typing.get_args(annotation)
        return Constrained.create(data_type=infer(args[0], feature), constraints=list(args[1:]))
    if origin is types.UnionType and annotation is Native:
        return Any
    if origin is typing.Any:
        return Any
    if issubclass(origin, type(None)):
        return Void
    if issubclass(origin, CustomDataType):
        return origin.attach(feature)
    if dataclasses.is_dataclass(origin):
        docstring = parse_docstring(origin, feature=feature)
        data_type = Structure.create(
            elements=docstring.parameters, name=origin.__name__, description=docstring.description
        )
        data_type._class = origin
        return data_type
    if issubclass(origin, DataType):
        return origin
    if issubclass(origin, enum.Enum):
        return Enum.create(origin)
    if issubclass(origin, bool):
        return Boolean
    if issubclass(origin, int):
        return Integer
    if issubclass(origin, float):
        return Real
    if issubclass(origin, str):
        return String
    if issubclass(origin, bytes):
        return Binary
    if issubclass(origin, datetime.datetime):
        return Timestamp
    if issubclass(origin, datetime.date):
        return Date
    if issubclass(origin, datetime.time):
        return Time
    if issubclass(origin, list):
        arg = typing.get_args(annotation)
        if not arg:
            msg = f"Unable to identify SiLA type from annotation '{annotation}'"
            raise TypeError(msg)

        return List.create(data_type=infer(arg[0], feature))
    if issubclass(origin, collections.abc.AsyncIterable):
        arg = typing.get_args(annotation)
        if not arg:
            msg = f"Unable to identify SiLA type from annotation '{annotation}'"
            raise TypeError(msg)

        return infer(arg[0], feature)
    if inspect.isclass(origin):
        docstring = parse_docstring(origin, feature=feature)
        data_type = Structure.create(
            elements=docstring.parameters, name=origin.__name__, description=docstring.description
        )
        data_type._class = origin
        return data_type

    msg = f"Unable to identify SiLA type from annotation '{annotation}'"
    raise TypeError(msg)



import dataclasses
import typing

from sila import Any, Binary, Boolean, Date, Element, Integer, List, Real, String, Structure, Time, Timestamp, datetime
from unitelabs.cdk.sila.data_types.convert_data_type import to_sila


class TestToSila:
    def test_should_convert_string(self):
        value = "Hello, World!"
        responses = {"string_value": Element("StringValue", "String Value", data_type=String)}

        response = to_sila(value, responses)

        assert response == {"string_value": "Hello, World!"}

    def test_should_convert_binary(self):
        value = b"Hello, World!"
        responses = {"binary_value": Element("BinaryValue", "Binary Value", data_type=Binary)}

        response = to_sila(value, responses)

        assert response == {"binary_value": b"Hello, World!"}

    def test_should_convert_boolean(self):
        value = True
        responses = {"boolean_value": Element("BooleanValue", "Boolean Value", data_type=Boolean)}

        response = to_sila(value, responses)

        assert response == {"boolean_value": True}

    def test_should_convert_integer(self):
        value = 42
        responses = {"integer_value": Element("IntegerValue", "Integer Value", data_type=Integer)}

        response = to_sila(value, responses)

        assert response == {"integer_value": 42}

    def test_should_convert_real(self):
        value = 3.141592653589793
        responses = {"real_value": Element("RealValue", "Real Value", data_type=Real)}

        response = to_sila(value, responses)

        assert response == {"real_value": 3.141592653589793}

    def test_should_convert_date(self):
        value = datetime.date(
            year=2022, month=8, day=5, tzinfo=datetime.timezone(datetime.timedelta(hours=12, minutes=34))
        )
        responses = {"date_value": Element("DateValue", "Date Value", data_type=Date)}

        response = to_sila(value, responses)

        assert response == {
            "date_value": datetime.date(
                year=2022, month=8, day=5, tzinfo=datetime.timezone(datetime.timedelta(hours=12, minutes=34))
            )
        }

    def test_should_convert_time(self):
        value = datetime.time(
            hour=12,
            minute=34,
            second=56,
            microsecond=789000,
            tzinfo=datetime.timezone(datetime.timedelta(hours=12, minutes=34)),
        )
        responses = {"time_value": Element("TimeValue", "Time Value", data_type=Time)}

        response = to_sila(value, responses)

        assert response == {
            "time_value": datetime.time(
                hour=12,
                minute=34,
                second=56,
                microsecond=789000,
                tzinfo=datetime.timezone(datetime.timedelta(hours=12, minutes=34)),
            )
        }

    def test_should_convert_timestamp(self):
        value = datetime.datetime(
            year=2022,
            month=8,
            day=5,
            hour=12,
            minute=34,
            second=56,
            microsecond=789000,
            tzinfo=datetime.timezone(datetime.timedelta(hours=12, minutes=34)),
        )
        responses = {"timestamp_value": Element("TimestampValue", "Timestamp Value", data_type=Timestamp)}

        response = to_sila(value, responses)

        assert response == {
            "timestamp_value": datetime.datetime(
                year=2022,
                month=8,
                day=5,
                hour=12,
                minute=34,
                second=56,
                microsecond=789000,
                tzinfo=datetime.timezone(datetime.timedelta(hours=12, minutes=34)),
            )
        }

    def test_should_convert_any(self):
        value = "Hello, World!"
        responses = {"any_value": Element("AnyValue", "Any Value", data_type=Any)}

        response = to_sila(value, responses)

        assert response == {"any_value": "Hello, World!"}

    def test_should_convert_list(self):
        value = ["SiLA 2", "is", "awesome!"]
        responses = {"list_value": Element("ListValue", "List Value", data_type=List)}

        response = to_sila(value, responses)

        assert response == {"list_value": ["SiLA 2", "is", "awesome!"]}

    def test_should_convert_tuple_to_structure(self):
        value = ("Hello, World!", 42)
        responses = {
            "string_value": Element("StringValue", "String Value", data_type=String),
            "integer_value": Element("IntegerValue", "Integer Value", data_type=Integer),
        }

        response = to_sila(value, responses)

        assert response == {"string_value": "Hello, World!", "integer_value": 42}

    def test_should_convert_dict_to_structure(self):
        value = {"string_value": "Hello, World!", "integer_value": 42}
        responses = {
            "string_value": Element("StringValue", "String Value", data_type=String),
            "integer_value": Element("IntegerValue", "Integer Value", data_type=Integer),
        }

        response = to_sila(value, responses)

        assert response == {"string_value": "Hello, World!", "integer_value": 42}

    def test_should_convert_dict_to_nested_structure(self):
        value = {"string_value": "Hello, World!", "integer_value": 42}
        responses = {
            "structure_value": Element(
                "StructureValue",
                "Structure Value",
                data_type=Structure.create(
                    {
                        "string_value": Element("StringValue", "String Value", data_type=String),
                        "integer_value": Element("IntegerValue", "Integer Value", data_type=Integer),
                    }
                ),
            )
        }

        response = to_sila(value, responses)

        assert response == {"structure_value": {"string_value": "Hello, World!", "integer_value": 42}}

    def test_should_convert_typed_dict_to_nested_structure(self):
        class TypedDict(typing.TypedDict):
            string_value: str
            integer_value: int

        value = TypedDict(string_value="Hello, World!", integer_value=42)
        responses = {
            "structure_value": Element(
                "StructureValue",
                "Structure Value",
                data_type=Structure.create(
                    {
                        "string_value": Element("StringValue", "String Value", data_type=String),
                        "integer_value": Element("IntegerValue", "Integer Value", data_type=Integer),
                    }
                ),
            )
        }

        response = to_sila(value, responses)

        assert response == {"structure_value": {"string_value": "Hello, World!", "integer_value": 42}}

    def test_should_convert_class_to_nested_structure(self):
        @dataclasses.dataclass
        class Dataclass:
            string_value: str
            integer_value: int

        value = Dataclass(string_value="Hello, World!", integer_value=42)
        responses = {
            "structure_value": Element(
                "StructureValue",
                "Structure Value",
                data_type=Structure.create(
                    {
                        "string_value": Element("StringValue", "String Value", data_type=String),
                        "integer_value": Element("IntegerValue", "Integer Value", data_type=Integer),
                    }
                ),
            )
        }

        response = to_sila(value, responses)

        assert response == {"structure_value": Dataclass(string_value="Hello, World!", integer_value=42)}



import typing
import unittest.mock

import pytest
from sila.framework.errors.defined_execution_error import DefinedExecutionError

import sila
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.data_types.any import Any
from unitelabs.cdk.sila.property.unobservable_property import UnobservableProperty


class DefinedError(Exception):
    pass


@pytest.fixture
def feature():
    return Feature(identifier="Feature", name="Feature")


@pytest.fixture
def handler():
    return UnobservableProperty(identifier="UnobservableProperty", name="Unobservable Property", errors=[DefinedError])


class TestAttach:
    async def test_should_attach_handler(self, feature: Feature, handler: UnobservableProperty):
        # Attach
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is True
        assert feature.properties[handler._identifier] == handler._handler

    async def test_should_ignore_disabled_handler(self, feature: Feature, handler: UnobservableProperty):
        # Attach
        handler._enabled = False
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is False
        assert handler._identifier not in feature.properties

    async def test_should_call_enabled_callback(self, feature: Feature, handler: UnobservableProperty):
        # Attach
        handler._enabled = unittest.mock.Mock(return_value=True)
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is True
        assert feature.properties[handler._identifier] == handler._handler
        handler._enabled.assert_called_once_with(feature)

    async def test_should_attach_return_annotation(self, feature: Feature, handler: UnobservableProperty):
        # Create handler
        def function() -> int:
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.UnobservableProperty)
        assert handler._handler.data_type == sila.Integer

    async def test_should_attach_missing_return_annotation(self, feature: Feature, handler: UnobservableProperty):
        # Create handler
        def function():
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.UnobservableProperty)
        assert handler._handler.data_type == Any

    async def test_should_attach_annotated_return(self, feature: Feature, handler: UnobservableProperty):
        # Create handler
        def function() -> typing.Annotated[int, sila.Unit("s", [sila.UnitComponent("Second")])]:
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.UnobservableProperty)
        assert issubclass(handler._handler.data_type, sila.Constrained)
        assert handler._handler.data_type.data_type == sila.Integer
        assert handler._handler.data_type.constraints == [sila.Unit("s", [sila.UnitComponent("Second")])]

    async def test_should_attach_documented_error(self, feature: Feature, handler: UnobservableProperty):
        # Create handler
        def function() -> int:
            """
            A function that returns an integer.

            Raises:
              RuntimeError: When something goes wrong.
            """
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.UnobservableProperty)
        assert len(handler._handler.errors) == 2

        assert handler._handler.errors["DefinedError"].identifier == "DefinedError"
        assert handler._handler.errors["DefinedError"].display_name == "Defined Error"
        assert handler._handler.errors["DefinedError"].description == "Common base class for all non-exit exceptions."

        assert handler._handler.errors["RuntimeError"].identifier == "RuntimeError"
        assert handler._handler.errors["RuntimeError"].display_name == "Runtime Error"
        assert handler._handler.errors["RuntimeError"].description == "When something goes wrong."

    async def test_should_recognize_duplicate_errors(self, feature: Feature, handler: UnobservableProperty):
        # Create handler
        def function() -> int:
            """
            A function that returns an integer.

            Raises:
              DefinedError: When something goes wrong.
            """
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.UnobservableProperty)
        assert len(handler._handler.errors) == 1
        assert handler._handler.errors["DefinedError"].identifier == "DefinedError"
        assert handler._handler.errors["DefinedError"].display_name == "Defined Error"
        assert handler._handler.errors["DefinedError"].description == "When something goes wrong."


class TestExecute:
    # Execute synchronous function with default parameters.
    async def test_execute_synchronous_default_parameters(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        callback = unittest.mock.Mock()

        def function():
            callback()
            return unittest.mock.sentinel.response

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={})

        # Assert that the function was called with the correct arguments
        callback.assert_called_once_with()

        # Assert that the method returns the correct value
        assert result == {"UnobservableProperty": unittest.mock.sentinel.response}

    # Execute synchronous function with no return value.
    async def test_execute_synchronous_returns_any(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        callback = unittest.mock.Mock()

        def function():
            callback()

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={})

        # Assert that the method returns the correct value
        callback.assert_called_once_with()
        assert result == {"UnobservableProperty": None}

    # Verify that the method raises an error when the synchronous function raises.
    async def test_raises_when_synchronous_raises(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        def function():
            msg = "Hello, World!"
            raise Exception(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(Exception, match=r"Exception: Hello, World!"):
            await handler.execute(metadata={})

    # Verify that the method raises a defined execution error when the synchronous decorator knows the error type.
    async def test_raises_when_synchronous_raises_known_error(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        def function():
            msg = "Hello, World!"
            raise DefinedError(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(DefinedExecutionError) as exc_info:
            await handler.execute(metadata={})

        assert exc_info.value.identifier == "DefinedError"
        assert exc_info.value.display_name == "Defined Error"
        assert exc_info.value.description == "Common base class for all non-exit exceptions."
        assert exc_info.value.message == "Hello, World!"

    # Execute asynchronous function with default parameters.
    async def test_execute_asynchronous_default_parameters(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        callback = unittest.mock.AsyncMock()

        async def function():
            await callback()
            return unittest.mock.sentinel.response

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={})

        # Assert that the function was called with the correct arguments
        callback.assert_awaited_once_with()

        # Assert that the method returns the correct value
        assert result == {"UnobservableProperty": unittest.mock.sentinel.response}

    # Execute asynchronous function with no return value.
    async def test_execute_asynchronous_returns_any(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        callback = unittest.mock.AsyncMock()

        async def function():
            await callback()

        handler(function)
        handler.attach(feature)

        # Execute function
        result = await handler.execute(metadata={})

        # Assert that the method returns the correct value
        callback.assert_awaited_once_with()
        assert result == {"UnobservableProperty": None}

    # Verify that the method raises an error when the asynchronous function raises.
    async def test_raises_when_asynchronous_raises(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        async def function():
            msg = "Hello, World!"
            raise Exception(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(Exception, match=r"Exception: Hello, World!"):
            await handler.execute(metadata={})

    # Verify that the method raises a defined execution error when the asynchronous decorator knows the error type.
    async def test_raises_when_asynchronous_raises_known_error(self, feature: Feature, handler: UnobservableProperty):
        # Initialize the unobservable property
        async def function():
            msg = "Hello, World!"
            raise DefinedError(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        with pytest.raises(DefinedExecutionError) as exc_info:
            await handler.execute(metadata={})

        assert exc_info.value.identifier == "DefinedError"
        assert exc_info.value.display_name == "Defined Error"
        assert exc_info.value.description == "Common base class for all non-exit exceptions."
        assert exc_info.value.message == "Hello, World!"



from sila.framework.constraints import (
    AllowedTypes,
    Constraint,
    ContentType,
    ContentTypeParameter,
    ElementCount,
    FullyQualifiedIdentifier,
    Identifier,
    Length,
    MaximalElementCount,
    MaximalLength,
    MinimalElementCount,
    MinimalLength,
    Pattern,
    Schema,
    SchemaType,
    SIUnit,
    Unit,
    UnitComponent,
)

from .maximal_exclusive import MaximalExclusive
from .maximal_inclusive import MaximalInclusive
from .minimal_exclusive import MinimalExclusive
from .minimal_inclusive import MinimalInclusive
from .set import Set

__all__ = [
    "AllowedTypes",
    "Constraint",
    "ContentType",
    "ContentTypeParameter",
    "ElementCount",
    "FullyQualifiedIdentifier",
    "Identifier",
    "Length",
    "MaximalElementCount",
    "MaximalExclusive",
    "MaximalInclusive",
    "MaximalLength",
    "MinimalElementCount",
    "MinimalExclusive",
    "MinimalInclusive",
    "MinimalLength",
    "Pattern",
    "SIUnit",
    "Schema",
    "SchemaType",
    "Set",
    "Unit",
    "UnitComponent",
]



import collections.abc
import dataclasses
import inspect

import typing_extensions as typing
from sila.server import FeatureIdentifier, Native, SiLAError, UndefinedExecutionError

import sila

from ..common import Dataclass
from ..common.errors import define_error
from ..data_types import Any

if typing.TYPE_CHECKING:
    from ..common import Decorator, Feature

T = typing.TypeVar("T", bound=Any)


@dataclasses.dataclass
class Metadatum(typing.Generic[T], Dataclass):
    """
    Define additional metadata that is used to extend existing features with reusable functionality.

    Attributes:
      feature: The feature this metadata is registered with.

    Examples:
      Define a metadatum on the consumer:
      >>> @dataclasses.dataclass
      ... class MyMetadata(sila.Metadatum, errors=[SomeError]):
      ...   param_a: str
      ...   param_b: int
      ...
      ... class MyMetadataProvider(sila.Feature):
      ...   def __init__(self):
      ...     super().__init__(metadata=[MyMetadata])
      ...
      ... class MyFeature(sila.Feature):
      ...   @sila.UnobservableProperty()
      ...   async def my_property(self, *, metadata: typing.Annotated[sila.Metadata, MyMetadata]) -> str:
      ...     \"\"\"Describe what your property does.\"\"\"
      ...     return metadata.param_a

      Define a metadatum on the provider:
      >>> @dataclasses.dataclass
      ... class MyMetadata(sila.Metadatum):
      ...   param: str
      ...
      ...   @typing.override
      ...   @classmethod
      ...   def affects(cls) -> list[sila.identifiers.FeatureIdentifier]:
      ...     return [cls.feature.app.get_feature(MyFeature).fully_qualified_identifier]
      ...
      ...   @typing.override
      ...   async def intercept(self, context: sila.Handler) -> None:
      ...     # Optionally, intercepts the execution of the property
      ...     # and aborts it if an error is raised
      ...     pass
      ...
      ... class MyMetadataProvider(sila.Feature):
      ...   def __init__(self):
      ...     super().__init__(metadata=[MyMetadata])
      ...
      ... class MyFeature(sila.Feature):
      ...   @sila.UnobservableProperty()
      ...   async def my_property(self) -> str:
      ...     \"\"\"Describe what your property does.\"\"\"
      ...     return ""
    """

    feature: typing.ClassVar["Feature"]
    _affects: typing.ClassVar[set[str]] = set()
    _metadatum: typing.ClassVar[type[sila.server.Metadata] | None] = None

    def __init_subclass__(
        cls,
        /,
        *,
        identifier: str | None = None,
        display_name: str | None = None,
        name: str | None = None,
        errors: collections.abc.Sequence[type[Exception]] | None = None,
    ) -> None:
        super().__init_subclass__(identifier=identifier, display_name=display_name, name=name)

        cls._affects = set()
        cls._errors: list[type[Exception]] = list(errors or [])

    @typing.override
    @classmethod
    def attach(cls, feature: "Feature") -> type[sila.server.Metadata]:
        super().attach(feature)
        cls.feature = feature

        if cls._identifier in feature.metadata:
            cls._affects.update(cls.affects())
            feature.metadata[cls._identifier].affects = list(cls._affects)
            cls._metadatum = typing.cast(type[sila.server.Metadata], feature.metadata[cls._identifier])

            return cls._metadatum

        data_type = cls._infer_data_type(feature)
        cls._metadatum = sila.server.Metadata.create(
            identifier=cls._identifier,
            display_name=cls._name,
            description=cls._description,
            errors={Error.identifier: Error for error in cls._errors if (Error := define_error(error))},
            data_type=data_type,
            affects=list(cls._affects),
            function=cls._intercept,
            feature=feature,
        )
        return cls._metadatum

    @classmethod
    async def _intercept(cls, value: Native, context: sila.Handler) -> None:
        """Intercept method execution."""

        try:
            await cls.from_native(value).intercept(context)
        except SiLAError:
            raise
        except Exception as error:
            if type(error) in cls._errors:
                raise define_error(error)(str(error)) from None

            msg = f"{error.__class__.__name__}: {error}"
            raise UndefinedExecutionError(msg) from error

    @classmethod
    def _infer_metadata(cls, decorator: "Decorator") -> tuple[str, list[type["Metadatum"]]]:
        from .metadata import Metadata

        signature = inspect.signature(decorator._function)

        parameter: str = ""
        metadata: list[type[Metadatum]] = []

        for param in signature.parameters.values():
            if (
                typing.get_origin(param.annotation) is typing.Annotated
                and (args := typing.get_args(param.annotation))
                and args[0] is Metadata
            ):
                parameter = param.name

                for arg in args[1:]:
                    if not issubclass(arg, Metadatum):
                        msg = (
                            f"Expected instance of `Metadatum` for metadata annotation, "
                            f"received '{arg.__name__}' for parameter '{param.name}'."
                        )
                        raise ValueError(msg)

                    if not decorator._handler:
                        raise RuntimeError

                    arg._affects.add(decorator._handler.fully_qualified_identifier)
                    if arg._metadatum:
                        arg._metadatum.affects = [
                            *arg._metadatum.affects,
                            decorator._handler.fully_qualified_identifier,
                        ]
                    metadata.append(arg)

                break

        return parameter, metadata

    @classmethod
    def from_native(cls, value: Native) -> typing.Self:
        """
        Convert a SiLA metadata value to this counterpart.

        Args:
          value: The value to parse.

        Returns:
          A new instance of this metadatum with the given value.
        """

        if cls._metadatum and issubclass(cls._metadatum.data_type, sila.Structure):
            return cls(**value)

        return cls(value)

    @classmethod
    def affects(self) -> list[FeatureIdentifier]:
        """Set the fully qualified identifiers of the handlers this metadata affects."""

        return []

    async def intercept(self, context: sila.Handler) -> None:
        """Intercept method execution."""



from .dataclass import Dataclass
from .decorator import Decorator
from .errors import DefinedExecutionError, define_error
from .feature import Feature

__all__ = ["Dataclass", "Decorator", "DefinedExecutionError", "Feature", "define_error"]



# ruff: noqa: D205, D401, D415, E501

import typing_extensions as typing

from unitelabs.cdk import sila


class AnyTypeTest(sila.Feature):
    """Provides commands and properties to set or respectively get SiLA Any Type values via command parameters or property responses respectively."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.UnobservableCommand()
    async def set_any_type_value(
        self, any_type_value: sila.Any
    ) -> tuple[
        typing.Annotated[
            str,
            sila.constraints.Schema(
                "Xml",
                url="https://gitlab.com/SiLA2/sila_base/-/raw/beb5d703ab62b1242695f3a591f04a07ebc7b528/schema/AnyTypeDataType.xsd",
            ),
        ],
        sila.Any,
    ]:
        """
        Receives an Any type value and returns the type and the value that has been received.

        Args:
          AnyTypeValue: The Any type value to be set.

        Returns:
          ReceivedAnyType: The type that has been received.
          ReceivedValue: The value that has been received.
        """

        if isinstance(any_type_value, dict):
            Structure = sila.data_types.Structure.create(
                {
                    "StringTypeValue": sila.data_types.Element(
                        identifier="StringTypeValue",
                        display_name="String Type Value",
                        description="A string value.",
                        data_type=sila.data_types.String,
                    ),
                    "IntegerTypeValue": sila.data_types.Element(
                        identifier="IntegerTypeValue",
                        display_name="Integer Type Value",
                        description="An integer value.",
                        data_type=sila.data_types.Integer,
                    ),
                    "DateTypeValue": sila.data_types.Element(
                        identifier="DateTypeValue",
                        display_name="Date Type Value",
                        description="A date value.",
                        data_type=sila.data_types.Date,
                    ),
                }
            )
            structure = await Structure.from_native(
                self.context,
                any_type_value,
            )
            any_type = sila.data_types.AnyType(structure)

            return any_type.schema, any_type

        data_type = await sila.data_types.AnyType.from_native(self.context, any_type_value)

        return data_type.schema, any_type_value

    @sila.UnobservableProperty()
    def get_any_type_string_value(self) -> sila.Any:
        """Returns the Any type String value 'SiLA_Any_type_of_String_type'."""

        return "SiLA_Any_type_of_String_type"

    @sila.UnobservableProperty()
    def get_any_type_integer_value(self) -> sila.Any:
        """Returns the Any type Integer value 5124."""

        return 5124

    @sila.UnobservableProperty()
    def get_any_type_real_value(self) -> sila.Any:
        """Returns an Any type Real value 3.1415926."""

        return 3.1415926

    @sila.UnobservableProperty()
    def get_any_type_boolean_value(self) -> sila.Any:
        """Returns the Any type Boolean value true."""

        return True

    @sila.UnobservableProperty()
    def any_type_binary_value(self) -> sila.Any:
        """Returns the Any type ASCII-encoded string value 'SiLA_Any_type_of_Binary_type' as Binary."""

        return b"SiLA_Any_type_of_Binary_type"

    @sila.UnobservableProperty()
    def get_any_type_date_value(self) -> sila.Any:
        """Returns the Any type Date value 05.08.2022 respective 08/05/2022, timezone +2."""

        return sila.datetime.date(
            year=2022, month=8, day=5, tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2))
        )

    @sila.UnobservableProperty()
    def get_any_type_time_value(self) -> sila.Any:
        """Returns the Any type Time value 12:34:56.789, timezone +2."""

        return sila.datetime.time(
            hour=12,
            minute=34,
            second=56,
            microsecond=789000,
            tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
        )

    @sila.UnobservableProperty()
    def get_any_type_timestamp_value(self) -> sila.Any:
        """Returns the Any type Timestamp value 2022-08-05 12:34:56.789, timezone +2."""

        return sila.datetime.datetime(
            year=2022,
            month=8,
            day=5,
            hour=12,
            minute=34,
            second=56,
            microsecond=789000,
            tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
        )

    @sila.UnobservableProperty()
    def any_type_list_value(self) -> sila.Any:
        """Returns the Any type String List value ('SiLA 2', 'Any', 'Type', 'String', 'List')"""

        return ["SiLA 2", "Any", "Type", "String", "List"]

    @sila.UnobservableProperty()
    async def any_type_structure_value(self) -> sila.Any:
        """
        Returns the following Any type Structure value:
        - String 'StringTypeValue' = 'A String value',
        - Integer 'IntegerTypeValue' = 83737665,
        - Date 'DateTypeValue' = 05.08.2022 respective 08/05/2022 timezone +2 )
        """

        Structure = sila.data_types.Structure.create(
            {
                "StringTypeValue": sila.data_types.Element(
                    identifier="StringTypeValue",
                    display_name="StringTypeValue",
                    description="Astringvalue.",
                    data_type=sila.data_types.String,
                ),
                "IntegerTypeValue": sila.data_types.Element(
                    identifier="IntegerTypeValue",
                    display_name="IntegerTypeValue",
                    description="Anintegervalue.",
                    data_type=sila.data_types.Integer,
                ),
                "DateTypeValue": sila.data_types.Element(
                    identifier="DateTypeValue",
                    display_name="DateTypeValue",
                    description="Adatevalue.",
                    data_type=sila.data_types.Date,
                ),
            }
        )

        return await Structure.from_native(
            self.context,
            {
                "StringTypeValue": "A String value",
                "IntegerTypeValue": 83737665,
                "DateTypeValue": sila.datetime.date(
                    year=2022, month=8, day=5, tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2))
                ),
            },
        )



import collections.abc
import dataclasses
import functools
import inspect
import warnings

import typing_extensions as typing
from sila.server import Server

import sila

from ..metadata import Metadatum
from ..utils import to_display_name, to_identifier
from .decorator import Decorator

if typing.TYPE_CHECKING:
    from ...connector import Connector


@dataclasses.dataclass
class Feature(sila.framework.Feature):
    """
    A feature describes a specific behavior of the server.

    Use the docstring of your feature class to provide a detailed,
    human-readable description of the use of your feature.
    """

    def __init__(
        self,
        *args,
        identifier: str | None = None,
        display_name: str | None = None,
        name: str | None = None,
        description: str | None = None,
        metadata: collections.abc.Sequence[type[Metadatum]] | None = None,
        **kwargs,
    ):
        if display_name is not None:
            msg = "Using `display_name` is deprecated, use `name` instead."
            warnings.warn(msg, stacklevel=2)
            name = display_name

        name = name or to_display_name(self.__class__.__name__)
        identifier = identifier or to_identifier(name)
        description = description or next((inspect.getdoc(cls) for cls in inspect.getmro(type(self))), "") or ""

        super().__init__(*args, identifier=identifier, display_name=name, description=description, **kwargs)

        self._metadata: dict[str, type[Metadatum]] = {}
        for metadatum in metadata or []:
            self._metadata[metadatum._identifier] = metadatum
            metadatum.attach(self).add_to_feature(self)

        self._handlers: dict[str, Decorator] = {}
        self._app: Connector | None = None

    def attach(self) -> bool:
        """
        Attach all handlers to this feature.

        Returns:
          Whether at least one handler was attached.
        """

        attached = False
        abc_methods = {}
        for m in inspect.getmro(type(self)):
            if "unitelabs.cdk.features" in m.__module__:
                abc_methods = dict(inspect.getmembers(m, predicate=inspect.isfunction))
                break
        if abc_methods:
            for name, function in inspect.getmembers(type(self), predicate=inspect.isfunction):
                # move handlers from ABCs to overrides
                if hasattr(function, "__override__") and getattr(function, "__override__", False):
                    abc_method = abc_methods.get(name)
                    if abc_method and (handler := getattr(abc_method, "__handler", None)):
                        setattr(function, "__handler", handler)

        for name, function in inspect.getmembers(type(self), predicate=inspect.isfunction):
            if (handler := getattr(function, "__handler", None)) and isinstance(handler, Decorator):
                handler = handler.clone()

                method = getattr(self, name).__func__
                method = functools.partial(method, self)
                handler._function = functools.wraps(function)(method)

                attached = handler.attach(self) or attached

                if handler._identifier in self._handlers:
                    prev_func = self._handlers[handler._identifier]._function

                    msg = (
                        f"Duplicate handler identifier '{handler._identifier}' detected for feature "
                        f"'{self.__class__.__name__}'. "
                        f"Existing: {self.__class__.__name__}.{prev_func.__name__}. "
                        f"New: {self.__class__.__name__}.{name} (will override). "
                        "To avoid unintended overrides, set a unique 'identifier' in your decorator or "
                        "rename one of the methods."
                    )
                    warnings.warn_explicit(
                        msg,
                        category=UserWarning,
                        filename=inspect.getfile(inspect.unwrap(function)),
                        lineno=inspect.getsourcelines(inspect.unwrap(function))[1],
                    )

                self._handlers[handler._identifier] = handler

        return attached or bool(self.metadata)

    def optimize(self) -> None:
        """Optimize the feature."""

        for metadata in self.metadata.values():
            affected_features: dict[str, Feature] = {}
            for affects in metadata.affects:
                if (affected_feature := self.server.get_feature(affects)) and isinstance(affected_feature, Feature):
                    affected_features[affected_feature.fully_qualified_identifier] = affected_feature

            affects = set(metadata.affects)
            for feature in affected_features.values():
                all_handlers = {
                    handler._handler.fully_qualified_identifier
                    for handler in feature._handlers.values()
                    if handler._handler
                }

                if all_handlers.issubset(affects):
                    affects.difference_update(all_handlers)
                    affects.add(feature.fully_qualified_identifier)

                metadata.affects = list(affects)

    @property
    def app(self) -> "Connector":
        """The connector app this feature is registered with."""

        if not self._app:
            raise RuntimeError

        return self._app

    @property
    def server(self) -> Server:
        """The server this feature is registered with."""

        if not isinstance(self.context, Server):
            raise RuntimeError

        return self.context



import asyncio
import collections.abc
import contextlib
import inspect
import logging
import weakref

import typing_extensions as typing

from sila import server

from . import sila
from .config import ConnectorBaseConfig
from .features.core.sila_service import SiLAService

Handler = collections.abc.Callable[..., typing.Any | typing.Awaitable[typing.Any]]

T = typing.TypeVar("T", bound=sila.Feature)


class Connector:
    """Main app."""

    def __init__(self, config: ConnectorBaseConfig | None = None):
        self.__config = config or ConnectorBaseConfig()

        self._ready = asyncio.Event()
        self._shutdown = asyncio.Event()
        self._shutdown_handlers: list[Handler] = []

        sila_server = self.config.sila_server
        if sila_server is None:
            self._sila_server = None
            self._discovery = None
        else:
            self._sila_server = server.Server(sila_server)
            if self.config.discovery is not None:
                self._discovery = server.Discovery(self._sila_server)
            else:
                self._discovery = None

        cloud_server_endpoint = self.config.cloud_server_endpoint
        if cloud_server_endpoint is None:
            self._cloud_server = None
        else:
            options = {
                # The period (in milliseconds) after which a keepalive ping is sent on the transport.
                "grpc.keepalive_time_ms": 60 * 1000,
                # The amount of time (in milliseconds) the sender of the keepalive ping waits for an
                # acknowledgement. If it does not receive an acknowledgment within this time, it will
                # close the connection.
                "grpc.keepalive_timeout_ms": 3 * 60 * 1000,
                # If set to 1 (0 : false; 1 : true), allows keepalive pings to be sent even if there
                # are no calls in flight.
                "grpc.keepalive_permit_without_calls": 0,
                # How many pings can the client send before needing to send a data/header frame.
                "grpc.http2.max_pings_without_data": 0,
            } | cloud_server_endpoint.options
            cloud_server_endpoint.options = options
            self._cloud_server = server.CloudServer(cloud_server_endpoint)
            if self._sila_server is not None:
                self._cloud_server.context = self._sila_server

        if self._sila_server is not None:
            self.register(SiLAService())

    async def start(self) -> None:
        """Start the connector and all related services."""
        tasks = []
        if self._sila_server:
            tasks.append(self._sila_server.start())
        if self._cloud_server:
            tasks.append(self._cloud_server.start())
        if self._discovery:
            tasks.append(self._discovery.start())

        await asyncio.gather(*tasks)
        self._ready.set()

    async def stop(self) -> None:
        """Stop the connector and all related services."""

        for shutdown_handler in self._shutdown_handlers:
            with contextlib.suppress(Exception):
                if inspect.iscoroutinefunction(shutdown_handler):
                    await shutdown_handler()
                else:
                    shutdown_handler()

        tasks = []
        if self._sila_server:
            tasks.append(self._sila_server.stop(grace=5))
        if self._cloud_server:
            tasks.append(self._cloud_server.stop(grace=5))
        if self._discovery:
            tasks.append(self._discovery.stop())

        await asyncio.gather(*tasks)

    async def wait_for_ready(self) -> None:
        """Wait until the connector is ready."""

        await self._ready.wait()
        self._ready.clear()

    async def wait_for_termination(self) -> None:
        """Wait until the connector is terminated."""

        await self._shutdown.wait()
        self._shutdown.clear()

    def get_feature(self, feature: type[T]) -> T:
        """
        Get the instance of a registered feature by its type.

        Args:
          feature: The type of the feature to receive.

        Returns:
          The feature registered with this connector.

        Raises:
          ValueError: If the given type is invalid or not
            recognized.
        """

        try:
            return next(feat for feat in self._sila_server.features.values() if isinstance(feat, feature))
        except StopIteration:
            msg = ""
            msg = f"Requested unknown feature '{feature.__name__}'."
            raise ValueError(msg) from None

    def register(self, feature: sila.Feature) -> None:
        """Register a new feature to this driver."""

        feature._app = weakref.proxy(self)
        if feature.attach():
            self.logger.debug("Added feature: %s", feature)
            if not self._sila_server:
                msg = "Cannot register feature when SiLA server is not configured."
                raise RuntimeError(msg)
            self._sila_server.register_feature(feature)
            feature.optimize()
        else:
            self.logger.debug("Skipped feature: %s", feature)

    @property
    def config(self) -> ConnectorBaseConfig:
        """The configuration."""

        return self.__config

    @property
    def sila_server(self) -> server.Server | None:
        """The SiLA Server."""
        return self._sila_server

    @property
    def logger(self) -> logging.Logger:
        """A standard Python :class:`~logging.Logger` for the app."""

        return logging.getLogger(__package__)

    @property
    def debug(self) -> bool:
        """Whether debug mode is enabled."""

        return True

    def on_shutdown(self, handler: Handler) -> None:
        """
        Add a shutdown hook to be called in the terminating phase.

        This will be in response to an explicit call to `app.stop()` or
        upon receipt of system signals such as SIGINT, SIGTERM or SIGHUP.

        Args:
          handler: The method to be called on shutdown.

        Raises:
          TypeError: If the `handler` argument is not callable.
        """

        if not callable(handler):
            msg = "The `handler` argument must be callable."
            raise TypeError(msg)

        self._shutdown_handlers.append(handler)

    def off_shutdown(self, handler: Handler) -> None:
        """
        Remove a previously added shutdown hook.

        Args:
          handler: The handler to be removed from the shutdown hooks.
        """

        with contextlib.suppress(ValueError):
            self._shutdown_handlers.remove(handler)



import asyncio
import collections.abc
import inspect
import time


def set_interval(function: collections.abc.Callable, delay: float = 1) -> asyncio.Task:
    """Repeatedly call a function or execute a codesnippet, with a fixed time delay between each call."""
    delay_ns = delay * 10**9
    timer = time.perf_counter_ns()

    async def interval(timer: float) -> None:
        while True:
            response = function()
            if inspect.isawaitable(response):
                await response

            timer += delay_ns
            await asyncio.sleep((timer - time.perf_counter_ns()) / 10**9)

    return asyncio.create_task(interval(timer))


def clear_interval(interval: asyncio.Task) -> None:
    """Cancel a timed, repeating action which was previously established by a call to set_interval()."""
    interval.cancel()



import dataclasses

from sila import Element, Integer, List, Real
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.data_types.structure import Structure


class TestFromNative:
    async def test_should_parse_structure_from_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestInnerStructure:
            value: int
            another_value: list[float]

        feature = Feature()
        data_type: type[Structure] = Structure.create(
            {
                "value": Element("Value", "Value", "", Integer),
                "another_value": Element("AnotherValue", "Another Value", "", List.create(Real)),
            }
        )
        data_type._class = TestInnerStructure

        # Convert from native
        value = await data_type.from_native(feature.context, {"value": 42, "another_value": [1.1, 2.2, 3.3]})

        # Assert that the method returns the correct value
        assert value == data_type(
            value={
                "value": Integer(value=42),
                "another_value": List.create(Real)(value=[Real(value=1.1), Real(value=2.2), Real(value=3.3)]),
            }
        )

    async def test_should_parse_structure_from_instance(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestInnerStructure:
            value: int
            another_value: list[float]

        feature = Feature()
        data_type: type[Structure] = Structure.create(
            {
                "value": Element("Value", "Value", "", Integer),
                "another_value": Element("AnotherValue", "Another Value", "", List.create(Real)),
            }
        )
        data_type._class = TestInnerStructure

        # Convert from native
        value = await data_type.from_native(
            feature.context, TestInnerStructure(value=42, another_value=[1.1, 2.2, 3.3])
        )

        # Assert that the method returns the correct value
        assert value == data_type(
            value={
                "value": Integer(value=42),
                "another_value": List.create(Real)(value=[Real(value=1.1), Real(value=2.2), Real(value=3.3)]),
            }
        )


class TestToNative:
    async def test_should_parse_structure(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestInnerStructure:
            value: int
            another_value: list[float]

        feature = Feature()
        data_type: type[Structure] = Structure.create(
            {
                "value": Element("Value", "Value", "", Integer),
                "another_value": Element("AnotherValue", "Another Value", "", List.create(Real)),
            }
        )
        data_type._class = TestInnerStructure
        value = await data_type.from_native(feature.context, {"value": 42, "another_value": [1.1, 2.2, 3.3]})

        # Convert to native
        native = await value.to_native(feature.context)

        # Assert that the method returns the correct value
        assert isinstance(native, TestInnerStructure)
        assert native.value == 42
        assert native.another_value == [1.1, 2.2, 3.3]



# ruff: noqa: D205, D401, D415, E501

import asyncio
import dataclasses
import datetime

import typing_extensions as typing

from unitelabs.cdk import sila


@dataclasses.dataclass
class String(sila.Metadatum):
    """A string"""

    string: str


class BinaryTransferTest(sila.Feature):
    """Provides commands and properties to set or respectively get the SiLA Basic Data Type Binary via command parameters or property responses respectively."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test", metadata=[String])

    @sila.UnobservableCommand()
    async def echo_binary_value(self, binary_value: bytes) -> bytes:
        """
        Receives a Binary value (transmitted either directly or via binary transfer) and
        returns the received value.

        Args:
          BinaryValue: The Binary value to be returned.

        Returns:
          ReceivedValue: The received Binary value transmitted in the same way it has been received.
        """

        return binary_value

    @sila.ObservableCommand()
    async def echo_binaries_observably(
        self, binaries: list[bytes], *, status: sila.Status, intermediate: sila.Intermediate[bytes]
    ) -> bytes:
        """
        Receives a list of binaries, echoes them individually as intermediate responses with a delay of 1 second, and then returns them as a single joint binary

        Args:
          Binaries: List of binaries to echo

        Yields:
          JointBinary: Single binary from the parameter list

        Returns:
          ReceivedValue: A single binary comprised of binaries received as parameter
        """

        for i, binary in enumerate(binaries):
            status.update(
                progress=(i + 1) / len(binaries),
                remaining_time=datetime.timedelta(seconds=len(binaries) - i),
            )
            intermediate.send(binary)

            await asyncio.sleep(1)

        return b"".join(binaries)

    @sila.UnobservableProperty()
    async def binary_value_directly(self) -> bytes:
        """Returns the UTF-8 encoded string 'SiLA2_Test_String_Value' directly transmitted as Binary value."""

        return b"SiLA2_Test_String_Value"

    @sila.UnobservableProperty()
    async def binary_value_download(self) -> bytes:
        """
        Returns the Binary Transfer UUID to be used to download the binary data which is the UTF-8 encoded string
        'A_slightly_longer_SiLA2_Test_String_Value_used_to_demonstrate_the_binary_download', repeated 100,000 times.
        """

        return b"A_slightly_longer_SiLA2_Test_String_Value_used_to_demonstrate_the_binary_download" * 100_000

    @sila.UnobservableCommand()
    async def echo_binary_and_metadata_string(
        self, binary: bytes, *, metadata: typing.Annotated[sila.Metadata, String]
    ) -> tuple[bytes, str]:
        """
        Receives a Binary and requires String Metadata, returns both

        Args:
          Binary: The binary to echo

        Returns:
          Binary: The received binary
          StringMetadata: The received String Metadata
        """

        return (binary, metadata[String].string)



from .docstring import Docstring, parse_docstring
from .interval import clear_interval, set_interval
from .name import to_display_name, to_identifier
from .version import SILA_SERVER_VERSION_PATTERN, parse_version

__all__ = [
    "SILA_SERVER_VERSION_PATTERN",
    "Docstring",
    "clear_interval",
    "parse_docstring",
    "parse_version",
    "set_interval",
    "to_display_name",
    "to_identifier",
]



import dataclasses
import sys

import sila

if sys.version_info < (3, 12):
    import typing_extensions as typing
else:
    import typing

T = typing.TypeVar("T", bound=sila.Native)


@dataclasses.dataclass
class Custom(typing.Generic[T], sila.Custom[T]):
    """Custom data type that converts from and to a python class instead of a dictionary."""

    _class: typing.ClassVar[type] = type(None)

    @typing.override
    @classmethod
    async def from_native(
        cls,
        context: "sila.Context",
        value: T | None = None,
        /,
        *,
        execution: sila.framework.common.Execution | None = None,
    ) -> typing.Self:
        if isinstance(value, cls._class):
            fields = dataclasses.fields(cls._class)
            value = {field.name: getattr(value, field.name) for field in fields}

            if not issubclass(cls.data_type, sila.Structure):
                field = fields[0]
                value = value.get(field.name)

        return await super().from_native(context, value, execution=execution)

    @typing.override
    async def to_native(self, context: "sila.Context", /) -> T:
        values = await super().to_native(context)

        if issubclass(self.data_type, sila.Structure):
            return self._class(**values)

        return self._class(values)



import abc

from unitelabs.cdk import sila


class GripControllerBase(sila.Feature, metaclass=abc.ABCMeta):
    """Control a robotic gripper for e.g. labware transfer operations."""

    def __init__(self):
        super().__init__(
            originator="io.unitelabs",
            category="robot",
            version="1.0",
            maturity_level="Draft",
        )

    @abc.abstractmethod
    @sila.UnobservableCommand()
    async def grip(self) -> None:
        """Close the gripper."""

    @abc.abstractmethod
    @sila.UnobservableCommand()
    async def release(self) -> None:
        """Release the gripper."""



import collections.abc

import typing_extensions as typing

from .metadatum import Metadatum

T = typing.TypeVar("T", bound=Metadatum)


class Metadata(collections.abc.Mapping):
    """Collection of the metadata sent by the client."""

    def __getitem__(self, key: type[T]) -> T:
        return super().__getitem__(key)  # type: ignore
